#
# (C) Copyright 2020 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
This file defines the ProviderRegistry class. The provider registry allows
user-defined providers to be created in the pywbem_mock environment and
registered with an instance of FakedWBEMConnection so that these user-defined
providers will be executed for their server requests in place of the default
server response behavior.  Providers are registered with the register_provider
method for particular namespaces, classnames servered by the provider and their
provider types (instance or method providers).

User defined providers can be registere for the following requests:

* CreateClass
* ModifyClass
* DeleteClass
* InvokeMethod
"""

import six
from pywbem._nocasedict import NocaseDict
from pywbem._utils import _format

from ._instancewriteprovider import InstanceWriteProvider
from ._methodprovider import MethodProvider
from ._utils import _uprint

__all__ = ['ProviderRegistry']


class ProviderRegistry(object):
    """
    This class defines the provider registry with methods to register
    a provider and to get the registered provider for a particular classname
    namespace, and provider_type.
    """
    #: Allowed provider types
    PROVIDER_TYPES = ["instance-write", "method"]

    def __init__(self):
        # Dictionary of registered providers.
        # Hierarchy of dictionaries is [namespace][classname][type]
        # value is provider object.
        self._registry = NocaseDict()

    def __repr__(self):
        return _format(
            "ProviderRegistry("
            "registry={s._registry}, ",
            s=self)

    def display_registered_providers(self, dest=None):
        """
        Generate display of registry in readable form and return the string
        Output format is:
        <namespace>:
          <classname>: CIM_Namespace provider: NamespaceProvider type: instance

        For example:

            Registered Providers:
            namespace: root/cimv2
              CIM_Foo      instance  UserInstanceTestProvider
              CIM_Foo      method    UserMethodTestProvider
            namespace: root/cimv3
              CIM_Foo      instance  UserInstanceTestProvider
              CIM_Foo      method    UserMethodTestProvider

        Parameters:

          dest (:term:`string`):
            File path of an output file. If `None`, the output is written to
            stdout.
        """
        def print_ljust(rows):
            """
            Print left justified and column aligned row of rows where each item
            is either a string or list of strings
            """

            widths = [max(map(len, col)) for col in zip(*rows)]
            for row in rows:
                _uprint(dest, u"  {0}".format(
                    u"  ".join((val.ljust(width) for
                                (val, width) in zip(row, widths)))))

        _uprint(dest, "Registered Providers:")
        for ns in self.provider_namespaces():
            rows = []
            _uprint(dest, _format(u'namespace: {0}', ns))
            for class_name in sorted(self.provider_classes(ns)):
                for type_ in sorted(self.provider_types(ns, class_name)):
                    provider = self.provider_obj(ns, class_name, type_)
                    provider_cn = provider.__class__.__name__
                    rows.append([class_name, type_, provider_cn])
            print_ljust(rows)

    def register_provider(self, conn, provider, namespaces=None,
                          schema_pragma_files=None, verbose=None):
        # pylint: disable=line-too-long
        """
        Register the provider object for specific namespaces and CIM classes.
        Registering a provider tells the FakedWBEMConnection that the provider
        implementation provided with this call as the `provider` parameter is
        to be executed as the request response method for the namespaces
        defined in the `namespaces` parameter, the provider type defined in the
        provider 'provider_type` attribute of the `provider` and the classes
        defined in the provider `provider_classnames` attribute of the
        `provider`.

        The provider registration process includes:

        1. Validation that the namespaces defined for the provider exist.
        2. Validation that the superclass of the provider is consistent with
           the `provider_type` attribute defined in the provider.
        3. Installation of any CIM classes defined by the provider
           (`provider_classnames` attribute) including installation of
           dependencies for these classes using the `schema_pragma_files` to
           locate the search directories for dependencies.
        4. Adding the provider to the registry of user_providers so that any
           of the request methods defined for the `provider_type` are
           passed to this provider in place of the default request processors.
        5. Execute post_register_setup() call to the provider to allow the
           provider to perform any special setup functionality.

        Providers can only be registered for the following request response
        methods:

        1. provider_type = 'instance-write': defines methods for CreateInstance,
           ModifyInstance, and DeleteInstance requests within a subclass of
           the `InstanceWriteProvider` class.

        2. provider_type = 'method': defines a InvokeMethod method within
           a subclass of the `MethodProvider` class.

        Each classname in a particular namespace may have at most one provider
        registered

        Parameters:

          conn  (:class:`~pywbem_mock.FakedWBEMConnection`):
            Defines the attributes of the connection. Used to issue requests to
            create instances in the Interop namespace for all existing
            namespaces that do not have instances of CIM_Namespace defined

          provider (instance of subclass of :class:`pywbem_mock.InstanceWriteProvider` or :class:`pywbem_mock.MethodProvider`):
            The methods in this subclass override the corresponding methods in
            the superclass. The method call parameters must be the same
            as the default method in the superclass and it must return
            data in the same format if the default method returns data.
             This
            class must contain variables `provider_type` and
            `provider_classnames` that define the type of provider and the CIM
            classes that the provider serves.

          namespaces (:term:`string` or :class:`py:list` of :term:`string`):
            Namespace or namespaces for which the provider is being registered.

            If `None`, the default namespace of the connection will be set to
            the built-in default namespace

          schema_pragma_files (:term:`py:iterable` of :term:`string` or :term:`string`):
            Path names of schema pragma files for the set of CIM
            classes that make up a schema such as the DMTF schema. These files
            must contain include pragams defining the file location of the
            classes to be compiled for the defined provider and for any
            dependencies required to compile those classes.  The directory
            containing each schema pragma file is passed to the MOF compiler as
            the search path for compile dependencies.

            see :class:`pywbem.MOFCompiler` for more information on the
            `search_paths` parameter.

          verbose (:class:`py:bool`):
            Display details on actions

        Raises:

            TypeError: Invalid provider_type retrieved from provider or
                       provider_type does not match superlclass. or the
                       namespace parameter is invalid.
            ValueError: provider_type retrieved from provider is not a
                        valid type string.
            ValueError: classnames parameter not a valid string or iterable or
                        namespace does not exist in repository.
        """  # noqa: E501
        # pylint: enable=line-too-long

        if schema_pragma_files:
            if isinstance(schema_pragma_files, six.string_types):
                schema_pragma_files = [schema_pragma_files]

        try:
            provider_type = provider.provider_type
            provider_classnames = provider.provider_classnames
        except AttributeError as ae:
            raise TypeError(
                _format("Attributes provider_type and provider_classnames "
                        "required in provider. exception {}",
                        ae))
        if provider_type == 'instance-write':
            if not isinstance(provider, InstanceWriteProvider):
                raise TypeError(
                    _format("Provider argument {0!A} is not a "
                            "valid subclass of InstanceWriteProvider. ",
                            provider))
        elif provider_type == 'method':
            if not isinstance(provider, MethodProvider):
                raise TypeError(
                    _format("Provider argument {0!A} is not a "
                            "valid subclass of MethodProvider. ",
                            provider))
        else:
            raise ValueError(
                _format("Provider_type argument {0!A} is not a valid provider "
                        "type. Valid provider types are {1!A}.",
                        provider_type, self.PROVIDER_TYPES))

        if provider_classnames is None:
            raise ValueError(
                _format('Classnames argument must be string '
                        'or list of strings. None not allowed.'))

        if namespaces is None:
            namespaces = [conn.default_namespace]

        if isinstance(namespaces, six.string_types):
            namespaces = [namespaces]

        if not isinstance(namespaces, (list, tuple)):
            raise TypeError(
                _format('Namespace argument {0|A} must be a string or '
                        'list/tuple but is {1}',
                        namespaces, type(namespaces)))

        for namespace in namespaces:
            if not isinstance(namespace, six.string_types):
                raise TypeError(
                    _format('Namespace "{0!A}" in namespaces argument not '
                            'a string. ', namespace))

            if namespace not in conn.namespaces:
                raise ValueError(
                    _format('Namespace "{0!A}" in namespaces argument not '
                            'in CIM repository. '
                            'Existing namespaces are: {1!A}. ',
                            namespace, conn.namespaces))

        if isinstance(provider_classnames, six.string_types):
            provider_classnames = [provider_classnames]
        assert isinstance(provider_classnames, (list, tuple))

        for classname in provider_classnames:
            assert isinstance(classname, six.string_types)

        # For each namespace in which the provider is to be registered,
        # if the class is not in that namespace, either compile it in if
        # pragmas exist or generate an exception if no pragmas exist

        for namespace in namespaces:
            for classname in provider_classnames:
                # pylint: disable=protected-access
                if not conn._mainprovider.class_exists(namespace, classname):
                    if schema_pragma_files:
                        conn.compile_schema_classes(
                            provider_classnames,
                            schema_pragma_files,
                            namespace=namespace,
                            verbose=verbose)
                    else:
                        raise ValueError(
                            _format('Class "{0!A}" does not exist in '
                                    'namespace {1!A} of the CIM repository '
                                    'and no schema pragma files were specified',
                                    classname, namespace))
            if namespace not in self._registry:
                self._registry[namespace] = NocaseDict()

            # Add classnames for the namespace
            for classname in provider_classnames:
                if classname not in self._registry[namespace]:
                    self._registry[namespace][classname] = {}
                self._registry[namespace][classname][provider_type] = provider

        if verbose:
            _format("Provider {0!A} registered: classes:[{1!A}],  "
                    "type: {1!A} namespaces:{2!A}",
                    provider.__class__.__name__,
                    ", ".join(provider_classnames),
                    provider_type, ", ".join(namespaces))

        # Call post_register_setup.  Since this method is defined in the
        # default provider methods (MethodProvider, etc.) any exception is
        # caught as an error.
        provider.post_register_setup(conn)

    def get_registered_provider(self, namespace, provider_type, classname):
        """
        Get the  user-defined provider registered for this namespace,
        provider_type, and classname.

        If no provider is registered, return `None`.

        Parameters:

          namespace (:term:`string`):
            The namespace in which the request will be executed.

          provider_type (:term:`string`):
            String containing keyword ('instance-write' or 'method') defining
            the type of provider.

          classname (:term:`string`):
            Name of the class defined for the operation

        Returns:

          Instance of :class:`~pywbem_mock.BaseProvider`: The registered
          provider.

          None if the registry is empty, the classname is not in the registry
          or the namespace is not in the registry or the entry for the
          classname, namespace and provider type is not defined.
        """

        if not self._registry or namespace not in self._registry:
            return None

        if classname not in self._registry[namespace]:
            return None

        # Return None if requested type is not registered type
        if provider_type not in self.provider_types(namespace, classname):
            return None

        return self.provider_obj(namespace, classname, provider_type)

    def provider_namespaces(self):
        """
        Get list of namespaces for registered providers. The returned
        list is case sensitive.

        Returns:
            :class:`py:list` of :term:`string`: namespaces
            for which providers are registered.
        """
        return list(self._registry.keys())

    def provider_classes(self, namespace):
        """
        Get case insensitive iterable  of the classes for providers registered
        for a namespace. The returned list is case sensitive.

        Parameters:

          namespace (:term:`string`):
            The namespace in which the request will be executed.

        Returns:
            :class:`py:list` of :term:`string`:
            Names of classes registered for namespace

        Raises:
          KeyError: if namespace invalid
        """

        return list(self._registry[namespace].keys())

    def provider_types(self, namespace, classname):
        """
        Get provider types for a namespace and classname.  This is
        a case-sensitive list.

        Parameters:

          namespace (:term:`string`):
            The namespace in which the request will be executed.

          classname (:term:`string`):
            Name of the class defined for the operation.

        Returns:
            :class:`py:list` of :term:`string`: Strings defining provider
            types for the defined namespace and classname

        Raises:
          KeyError: if namespace invalid
        """
        return list(self._registry[namespace][classname].keys())

    def provider_obj(self, namespace, classname, provider_type):
        """
        Get the registered provider object (instance of the registered
        provider) for namespace, provider classname, and provider_type.

        Parameters:

          namespace (:term:`string`):
            The namespace in which the request will be executed.

          classname (:term:`string`):
            Name of the class defined for the operation.

          provider_type (:term:`string`):
            String containing keyword ('instance-write' or 'method') defining
            the type of provider.

        Returns:
            The registered object

        Raises:
          KeyError: if namespace or classname invalid
        """
        return self._registry[namespace][classname][provider_type]

    def iteritems(self):
        """
        Return an iterator through the registered provider items. Each item
        is a tuple(namespace, classname, provider_type, provider_obj),
        with:

          namespace (:term:`string`):
            The namespace in which the request will be executed.

          classname (:term:`string`):
            Name of the class defined for the operation.

          provider_type (:term:`string`):
            String containing keyword ('instance-write' or 'method') defining
            the type of provider.

          provider_obj (:class:`~pywbem_mock.BaseProvider`):
            The registered provider.
        """
        ns_dict = self._registry
        for ns in ns_dict:
            cln_dict = ns_dict[ns]
            for cln in cln_dict:
                pt_dict = cln_dict[cln]
                for pt in pt_dict:
                    pobj = pt_dict[pt]
                    yield (ns, cln, pt, pobj)

    def load(self, other):
        """
        Replace the data in this object with the data from the other object.

        This is used to restore the object from a serialized state, without
        changing its identity.
        """
        # pylint: disable=protected-access
        self._registry = other._registry



"""
Unit test cases for _cim_http.py
"""

from __future__ import absolute_import

import re
import requests
import urllib3
import pytest

from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import _cim_http  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


# These defaults are defined separately from those in _cim_constants.py to
# ensure that changes of the defaults are caught.
DEFAULT_PORT_HTTP = 5988
DEFAULT_PORT_HTTPS = 5989
DEFAULT_SCHEME = 'http'


TESTCASES_PARSE_URL = [

    # Testcases for _cim_http.parse_url()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * func_kwargs: input arguments to the function that is tested
    #   * exp_scheme: expected returned scheme, or None
    #   * exp_hostport: expected returned hostname, or None
    #   * exp_url: expected returned url, or None
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Default for allow_defaults argument is True",
        dict(
            func_kwargs=dict(
                url='myhost:5000',
                # allow_defaults defaults to True
            ),
            exp_scheme='http',
            exp_hostport='myhost:5000',
            exp_url='http://myhost:5000',
        ),
        None, None, True
    ),
    (
        "Invalid URL: empty string",
        dict(
            func_kwargs=dict(
                url='',
                allow_defaults=True,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Invalid URL: slash",
        dict(
            func_kwargs=dict(
                url='/',
                allow_defaults=True,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Applying default scheme raises ValueError when not allowed",
        dict(
            func_kwargs=dict(
                url='myhost:5000',
                allow_defaults=False,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Applying default scheme succeeds when allowed",
        dict(
            func_kwargs=dict(
                url='myhost:5000',
                allow_defaults=True,
            ),
            exp_scheme='{scheme}'.format(scheme=DEFAULT_SCHEME),
            exp_hostport='myhost:5000',
            exp_url='{scheme}://myhost:5000'.format(scheme=DEFAULT_SCHEME),
        ),
        None, None, True
    ),
    (
        "Applying default port for HTTPS raises ValueError when not allowed",
        dict(
            func_kwargs=dict(
                url='https://myhost',
                allow_defaults=False,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Applying default port for HTTPS succeeds when allowed",
        dict(
            func_kwargs=dict(
                url='https://myhost',
                allow_defaults=True,
            ),
            exp_scheme='https',
            exp_hostport='myhost:{port}'.format(port=DEFAULT_PORT_HTTPS),
            exp_url='https://myhost:{port}'.format(port=DEFAULT_PORT_HTTPS),
        ),
        None, None, True
    ),
    (
        "Applying default port for HTTP raises ValueError when not allowed",
        dict(
            func_kwargs=dict(
                url='http://myhost',
                allow_defaults=False,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Applying default port for HTTP succeeds when allowed",
        dict(
            func_kwargs=dict(
                url='http://myhost',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='myhost:{port}'.format(port=DEFAULT_PORT_HTTP),
            exp_url='http://myhost:{port}'.format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "Unsupported scheme raises ValueError",
        dict(
            func_kwargs=dict(
                url='httpx://myhost',
                allow_defaults=True,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Invalid port number raises ValueError",
        dict(
            func_kwargs=dict(
                url='http://myhost:123x',
                allow_defaults=True,
            ),
            exp_scheme=None,
            exp_hostport=None,
            exp_url=None,
        ),
        ValueError, None, True
    ),
    (
        "Hostname http without scheme or port",
        dict(
            func_kwargs=dict(
                url='http',
                allow_defaults=True,
            ),
            exp_scheme='{scheme}'.format(scheme=DEFAULT_SCHEME),
            exp_hostport='http:{port}'.format(port=DEFAULT_PORT_HTTP),
            exp_url='{scheme}://http:{port}'.
            format(scheme=DEFAULT_SCHEME, port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "Hostname https without scheme or port",
        dict(
            func_kwargs=dict(
                url='https',
                allow_defaults=True,
            ),
            exp_scheme='{scheme}'.format(scheme=DEFAULT_SCHEME),
            exp_hostport='https:{port}'.format(port=DEFAULT_PORT_HTTP),
            exp_url='{scheme}://https:{port}'.
            format(scheme=DEFAULT_SCHEME, port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "Hostname with port and trailing slash",
        dict(
            func_kwargs=dict(
                url='http://myhost:5000/',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='myhost:5000',
            exp_url='http://myhost:5000',
        ),
        None, None, True
    ),
    (
        "Hostname with port and trailing path segment",
        dict(
            func_kwargs=dict(
                url='http://myhost:5000/segment',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='myhost:5000',
            exp_url='http://myhost:5000',
        ),
        None, None, True
    ),
    (
        "Hostname without port and with trailing slash",
        dict(
            func_kwargs=dict(
                url='http://myhost/',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='myhost:{port}'.format(port=DEFAULT_PORT_HTTP),
            exp_url='http://myhost:{port}'.format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "Hostname without port and with trailing slash",
        dict(
            func_kwargs=dict(
                url='http://myhost/segment',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='myhost:{port}'.format(port=DEFAULT_PORT_HTTP),
            exp_url='http://myhost:{port}'.format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "Upper case scheme HTTP",
        dict(
            func_kwargs=dict(
                url='HTTP://myhost:5000',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='myhost:5000',
            exp_url='http://myhost:5000',
        ),
        None, None, True
    ),
    (
        "Upper case scheme HTTPS",
        dict(
            func_kwargs=dict(
                url='HTTPS://myhost:5000',
                allow_defaults=True,
            ),
            exp_scheme='https',
            exp_hostport='myhost:5000',
            exp_url='https://myhost:5000',
        ),
        None, None, True
    ),
    (
        "IPv4 address with port",
        dict(
            func_kwargs=dict(
                url='http://9.10.11.12',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='9.10.11.12:{port}'.format(port=DEFAULT_PORT_HTTP),
            exp_url='http://9.10.11.12:{port}'.format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format without scheme, zone index or port",
        dict(
            func_kwargs=dict(
                url='[2001:db8::7348]',
                allow_defaults=True,
            ),
            exp_scheme='{scheme}'.format(scheme=DEFAULT_SCHEME),
            exp_hostport='[2001:db8::7348]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
            exp_url='{scheme}://[2001:db8::7348]:{port}'.
            format(scheme=DEFAULT_SCHEME, port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format without zone index and without port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348]',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
            exp_url='http://[2001:db8::7348]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with numeric zone index delimited with % "
        "and without port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348%1]',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
            exp_url='http://[2001:db8::7348-1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with numeric zone index delimited with - "
        "and without port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348-1]',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
            exp_url='http://[2001:db8::7348-1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with string zone index delimited with % "
        "and without port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348%eth1]',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-eth1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
            exp_url='http://[2001:db8::7348-eth1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with string zone index delimited with - "
        "and without port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348-eth1]',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-eth1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
            exp_url='http://[2001:db8::7348-eth1]:{port}'.
            format(port=DEFAULT_PORT_HTTP),
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format without zone index and with port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348]:1234',
            exp_url='http://[2001:db8::7348]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with numeric zone index delimited with % "
        "and with port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348%1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-1]:1234',
            exp_url='http://[2001:db8::7348-1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with numeric zone index delimited with - "
        "and with port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348-1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-1]:1234',
            exp_url='http://[2001:db8::7348-1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with string zone index delimited with % "
        "and with port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348%eth1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-eth1]:1234',
            exp_url='http://[2001:db8::7348-eth1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with string zone index delimited with - "
        "and with port",
        dict(
            func_kwargs=dict(
                url='http://[2001:db8::7348-eth1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-eth1]:1234',
            exp_url='http://[2001:db8::7348-eth1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in (invalid but tolerated) text format without "
        "scheme and without zone index and with port",
        dict(
            func_kwargs=dict(
                url='2001:db8::7348:1234',
                allow_defaults=True,
            ),
            exp_scheme='{scheme}'.format(scheme=DEFAULT_SCHEME),
            exp_hostport='[2001:db8::7348]:1234',
            exp_url='{scheme}://[2001:db8::7348]:1234'.
            format(scheme=DEFAULT_SCHEME),
        ),
        None, None, True
    ),
    (
        "IPv6 address in (invalid but tolerated) text format "
        "without zone index and with port",
        dict(
            func_kwargs=dict(
                url='http://2001:db8::7348:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348]:1234',
            exp_url='http://[2001:db8::7348]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in (invalid but tolerated) text format "
        "with numeric zone index delimted with % and with port",
        dict(
            func_kwargs=dict(
                url='http://2001:db8::7348%1:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-1]:1234',
            exp_url='http://[2001:db8::7348-1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in (invalid but tolerated) text format "
        "with string zone index delimted with % and with port",
        dict(
            func_kwargs=dict(
                url='http://2001:db8::7348%eth1:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[2001:db8::7348-eth1]:1234',
            exp_url='http://[2001:db8::7348-eth1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with IPv4 portion, without zone index "
        "and with port",
        dict(
            func_kwargs=dict(
                url='http://[::ffff.9.10.11.12]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[::ffff.9.10.11.12]:1234',
            exp_url='http://[::ffff.9.10.11.12]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with IPv4 portion, with numeric zone index "
        "delimited with % and with port",
        dict(
            func_kwargs=dict(
                url='http://[::ffff.9.10.11.12%1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[::ffff.9.10.11.12-1]:1234',
            exp_url='http://[::ffff.9.10.11.12-1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with IPv4 portion, with numeric zone index "
        "delimited with - and with port",
        dict(
            func_kwargs=dict(
                url='http://[::ffff.9.10.11.12-1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[::ffff.9.10.11.12-1]:1234',
            exp_url='http://[::ffff.9.10.11.12-1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with IPv4 portion, with string zone index "
        "delimited with % and with port",
        dict(
            func_kwargs=dict(
                url='http://[::ffff.9.10.11.12%eth1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[::ffff.9.10.11.12-eth1]:1234',
            exp_url='http://[::ffff.9.10.11.12-eth1]:1234',
        ),
        None, None, True
    ),
    (
        "IPv6 address in URI format with IPv4 portion, with string zone index "
        "delimited with - and with port",
        dict(
            func_kwargs=dict(
                url='http://[::ffff.9.10.11.12-eth1]:1234',
                allow_defaults=True,
            ),
            exp_scheme='http',
            exp_hostport='[::ffff.9.10.11.12-eth1]:1234',
            exp_url='http://[::ffff.9.10.11.12-eth1]:1234',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_PARSE_URL)
@simplified_test_function
def test_parse_url(
        testcase, func_kwargs, exp_scheme, exp_hostport, exp_url):
    """
    Test function for _cim_http.parse_url()
    """

    # The code to be tested
    act_scheme, act_hostport, act_url = _cim_http.parse_url(**func_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the result
    assert act_scheme == exp_scheme
    assert act_hostport == exp_hostport
    assert act_url == exp_url


DUMMY_CONN_USING_HTTPS = pywbem.WBEMConnection('https://dummy')
DUMMY_CONN_USING_HTTP = pywbem.WBEMConnection('http://dummy')
DUMMY_POOL = urllib3.connectionpool.HTTPConnectionPool('dummy')


TESTCASES_REQUEST_EXC_MESSAGE = [

    # Testcases for _cim_http.request_exc_message()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * func_kwargs: input arguments to the function that is tested
    #   * exp_pattern: pattern for the expected message, or None
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Requests exception with no args",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException(),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^$',
        ),
        None, None, True
    ),
    (
        "Requests exception with empty message",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException(''),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^$',
        ),
        None, None, True
    ),
    (
        "Requests exception with simple message",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException('bla'),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^bla$',
        ),
        None, None, True
    ),
    (
        "Requests exception with simple message and HTTPS connection",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException('bla'),
                conn=DUMMY_CONN_USING_HTTPS,
            ),
            exp_pattern='^bla; OpenSSL version used: ',
        ),
        None, None, True
    ),
    (
        "Requests exception with attached MaxRetryError exception with "
        "underlying HTTPError exception",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException(
                    urllib3.exceptions.MaxRetryError(
                        pool=DUMMY_POOL,
                        url='http://dummy',
                        reason=urllib3.exceptions.HTTPError('bla'))),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^bla$',
        ),
        None, None, True
    ),
    (
        "Requests exception with attached HTTPError exception",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException(
                    urllib3.exceptions.HTTPError('bla')),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^bla$',
        ),
        None, None, True
    ),
    (
        "Requests exception with message type 1 that gets simplified",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException(
                    "(<foo>, 'bla')"
                ),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^bla$',
        ),
        None, None, True
    ),
    (
        "Requests exception with message type 2 that gets simplified",
        dict(
            func_kwargs=dict(
                exc=requests.exceptions.RequestException(
                    "<foo>: bla"
                ),
                conn=DUMMY_CONN_USING_HTTP,
            ),
            exp_pattern='^bla$',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_REQUEST_EXC_MESSAGE)
@simplified_test_function
def test_request_exc_message(testcase, func_kwargs, exp_pattern):
    """
    Test function for _cim_http.request_exc_message()
    """

    # The code to be tested
    act_message = _cim_http.request_exc_message(**func_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the result
    assert re.search(exp_pattern, act_message)


# TODO: Add unit tests for _cim_http.wbem_request(). It is already tested to
# some extent in the function tests, but adding unit tests would still be good.


TESTCASES_MAX_REPR = [

    # Testcases for _cim_http.max_repr()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * func_kwargs: input arguments to the function that is tested
    #   * exp_repr: expected repr result, or None
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Input text None",
        dict(
            func_kwargs=dict(
                text=None,
                max_len=20,
            ),
            exp_repr='None',
        ),
        None, None, True
    ),
    (
        "Input text shorter than max_len by much",
        dict(
            func_kwargs=dict(
                text='abc def',
                max_len=20,
            ),
            exp_repr="'abc def'",
        ),
        None, None, True
    ),
    (
        "Input text shorter than max_len by 1",
        dict(
            func_kwargs=dict(
                text='abc def',
                max_len=8,
            ),
            exp_repr="'abc def'",
        ),
        None, None, True
    ),
    (
        "Input text equal to max_len",
        dict(
            func_kwargs=dict(
                text='abc def',
                max_len=7,
            ),
            exp_repr="'abc def'",
        ),
        None, None, True
    ),
    (
        "Input text longer than max_len by 1",
        dict(
            func_kwargs=dict(
                text='abc def',
                max_len=6,
            ),
            exp_repr="'abc de'...",
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_MAX_REPR)
@simplified_test_function
def test_max_repr(testcase, func_kwargs, exp_repr):
    """
    Test function for _cim_http.max_repr()
    """

    # The code to be tested
    act_repr = _cim_http.max_repr(**func_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the result
    assert act_repr == exp_repr


TESTCASES_GET_CIMOBJECT_HEADER = [

    # Testcases for _cim_http.get_cimobject_header()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: input object for get_cimobject_header()
    #   * exp_result: expected result of get_cimobject_header(), or None
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Input object is text",
        dict(
            obj='abc',
            exp_result='abc',
        ),
        None, None, True
    ),
    (
        "Input object is CIMClassName without namespace",
        dict(
            obj=pywbem.CIMClassName('C1'),
            exp_result=':C1',
        ),
        None, None, True
    ),
    (
        "Input object is CIMClassName with namespace",
        dict(
            obj=pywbem.CIMClassName('C1', namespace='ns'),
            exp_result='ns:C1',
        ),
        None, None, True
    ),
    (
        "Input object is CIMInstanceName without namespace and no keys",
        dict(
            obj=pywbem.CIMInstanceName('C1'),
            exp_result=':C1',
        ),
        None, None, True
    ),
    (
        "Input object is CIMInstanceName with namespace and no keys",
        dict(
            obj=pywbem.CIMInstanceName('C1', namespace='ns'),
            exp_result='ns:C1',
        ),
        None, None, True
    ),
    (
        "Input object is CIMInstanceName with namespace and one key",
        dict(
            obj=pywbem.CIMInstanceName('C1', namespace='ns',
                                       keybindings=dict(K1='v1')),
            exp_result='ns:C1.K1="v1"',
        ),
        None, None, True
    ),
    (
        "Input object is CIMInstanceName with namespace and two keys",
        dict(
            obj=pywbem.CIMInstanceName(
                'C1', namespace='ns',
                keybindings=[('K1', 'v1'), ('K2', 'v2')]),
            exp_result='ns:C1.K1="v1",K2="v2"',
        ),
        None, None, True
    ),
    (
        "Input object has unsupported type CIMClass",
        dict(
            obj=pywbem.CIMClass('C1'),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Input object has unsupported type CIMInstance",
        dict(
            obj=pywbem.CIMInstance('C1'),
            exp_result=None,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_GET_CIMOBJECT_HEADER)
@simplified_test_function
def test_get_cimobject_header(testcase, obj, exp_result):
    """
    Test function for _cim_http.get_cimobject_header()
    """

    # The code to be tested
    act_result = _cim_http.get_cimobject_header(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the result
    assert act_result == exp_result



"""
Encapsulation of WBEM server definition file defining WBEM servers for pywbem
end2end tests.
"""

from __future__ import absolute_import

import os
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import errno
import yaml
import yamlloader

SDF_DIR = os.path.join('tests', 'server_definitions')

DEFAULT_SERVER_FILE = os.path.join(SDF_DIR, 'server_definition_file.yml')
EXAMPLE_SERVER_FILE = os.path.join(SDF_DIR,
                                   'example_server_definition_file.yml')


class ServerDefinitionFileError(Exception):
    """
    An error in the WBEM server definition file.
    """
    pass


class ServerDefinitionFile(object):
    """
    Encapsulation of the WBEM server definition file.
    """

    def __init__(self, filepath=DEFAULT_SERVER_FILE):
        self._filepath = filepath
        self._servers = OrderedDict()
        self._server_groups = OrderedDict()
        self._load_file()

    def _load_file(self):
        """Load the yaml file."""
        try:
            with open(self._filepath) as fp:
                try:
                    data = yaml.load(
                        fp, Loader=yamlloader.ordereddict.CSafeLoader)
                except (yaml.parser.ParserError,
                        yaml.scanner.ScannerError) as exc:
                    raise ServerDefinitionFileError(
                        "Invalid YAML syntax in WBEM server definition file "
                        "{0!r}: {1} {2}".
                        format(self._filepath, exc.__class__.__name__,
                               exc))
        except IOError as exc:
            if exc.errno == errno.ENOENT:
                raise ServerDefinitionFileError(
                    "The WBEM server definition file {0!r} was not found; "
                    "copy it from {1!r}".
                    format(self._filepath, EXAMPLE_SERVER_FILE))
            raise
        else:
            if data is None:
                raise ServerDefinitionFileError(
                    "The WBEM server definition file {0!r} is empty".
                    format(self._filepath))
            if not isinstance(data, (dict, OrderedDict)):
                # Starting with py38, this is a standard dict
                raise ServerDefinitionFileError(
                    "The WBEM server definition file {0!r} must contain a "
                    "dictionary at the top level, but contains {1}".
                    format(self._filepath, type(data)))

            if 'servers' not in data:
                raise ServerDefinitionFileError(
                    "The WBEM server definition file {0!r} does not define a "
                    "'servers' item, but items: {1}".
                    format(self._filepath, data.keys()))
            servers = data.get('servers')
            if not isinstance(servers, (dict, OrderedDict)):
                raise ServerDefinitionFileError(
                    "'servers' in WBEM server definition file {0!r} "
                    "must be a dictionary, but is a {1}".
                    format(self._filepath, type(servers)))
            self._servers.update(servers)

            server_groups = data.get('server_groups', OrderedDict())
            if not isinstance(server_groups, (dict, OrderedDict)):
                raise ServerDefinitionFileError(
                    "'server_groups' in WBEM server definition file {0!r} "
                    "must be a dictionary, but is a {1}".
                    format(self._filepath, type(server_groups)))
            for sg_nick in server_groups:
                visited_sg_nicks = list()
                self._check_sg(sg_nick, server_groups, servers,
                               visited_sg_nicks)
            self._server_groups.update(server_groups)

    def _check_sg(self, sg_nick, server_groups, servers, visited_sg_nicks):
        """Check server groups for nicknames"""
        visited_sg_nicks.append(sg_nick)
        server_group = server_groups[sg_nick]
        if not isinstance(server_group, list):
            raise ServerDefinitionFileError(
                "Server group {0!r} in WBEM server definition file {1!r} "
                "must be a list, but is a {2}".
                format(sg_nick, self._filepath, type(server_group)))
        for nick in server_group:
            if nick in visited_sg_nicks:
                raise ServerDefinitionFileError(
                    "Circular reference: Server group {0!r} in WBEM server "
                    "definition file {1!r} contains server group {2!r}, which "
                    "directly or indirectly contains server group {0!r}".
                    format(sg_nick, self._filepath, nick))
            if not isinstance(nick, str):
                raise ServerDefinitionFileError(
                    "Item {0!r} in server group {1!r} in WBEM server "
                    "definition file {2!r} must be a string, but is a {3}".
                    format(nick, sg_nick, self._filepath, type(nick)))
            if nick in server_groups:
                self._check_sg(nick, server_groups, servers, visited_sg_nicks)
            elif nick not in servers:
                raise ServerDefinitionFileError(
                    "Item {0!r} in server group {1!r} in WBEM server "
                    "definition file {2!r} is not a known server or server "
                    "group".
                    format(nick, sg_nick, self._filepath))

    @property
    def filename(self):
        """
        Path name of the WBEM server definition file.
        """
        return self._filepath

    def get_server(self, nickname):
        """
        Return a `ServerDefinition` object for the server with the specified
        nickname.
        """
        try:
            server_dict = self._servers[nickname]
        except KeyError:
            raise ValueError(
                "Server with nickname {0!r} not found in WBEM server "
                "definition file {1!r}".
                format(nickname, self._filepath))
        return ServerDefinition(nickname, server_dict)

    def list_servers(self, nickname):
        """
        Return a list of `ServerDefinition` objects for the servers in the
        server group with the specified nickname, or the single server with
        the specified nickname.
        """
        if nickname in self._servers:
            return [self.get_server(nickname)]

        if nickname in self._server_groups:  # pylint: disable no-else-return
            sd_list = list()  # of ServerDefinition objects
            sd_nick_list = list()  # of server nicknames
            for item_nick in self._server_groups[nickname]:
                for sd in self.list_servers(item_nick):
                    if sd.nickname not in sd_nick_list:
                        sd_list.append(sd)
                        sd_nick_list.append(sd.nickname)
            return sd_list

        raise ValueError(
            "Server group or server with nickname {0!r} not found in WBEM "
            "server definition file {1!r}".
            format(nickname, self._filepath))

    def list_all_servers(self):
        """
        Return a list of all servers in the WBEM server definition file.
        """
        return [self.get_server(nickname) for nickname in self._servers]


class ServerDefinition(object):
    """
    Encapsulation of a WBEM server definition (e.g. from a WBEM server
    definition file).
    """

    def __init__(self, nickname, server_dict):
        self._nickname = nickname
        self._description = server_dict.get('description', '')
        self._url = self._required_attr(server_dict, 'url', nickname)
        self._user = self._required_attr(server_dict, 'user', nickname)
        self._password = self._required_attr(server_dict, 'password', nickname)
        self._cert_file = server_dict.get('cert_file', None)
        self._key_file = server_dict.get('key_file', None)
        self._ca_certs = server_dict.get('ca_certs', None)
        self._no_verification = server_dict.get('no_verification', True)
        self._implementation_namespace = server_dict.get(
            'implementation_namespace', None)

    def _required_attr(self, server_dict, attr_name, nickname):
        # pylint: disable=no-self-use
        """Return the sever_dict attribute or a KeyError"""
        try:
            return server_dict[attr_name]
        except KeyError:
            raise ServerDefinitionFileError(
                "Required server attribute is missing in definition of server "
                "{0}: {1}".format(nickname, attr_name))

    def __repr__(self):
        return "ServerDefinition(" \
            "nickname={s.nickname!r}, " \
            "description={s.description!r}, " \
            "url={s.url!r}, " \
            "user={s.user!r}, " \
            "password=..., " \
            "cert_file={s.cert_file!r}, " \
            "key_file={s.key_file!r}, " \
            "ca_certs={s.ca_certs!r}, " \
            "no_verification={s.no_verification!r}) " \
            "implementation_namespace={s.implementation_namespace!r})". \
            format(s=self)

    @property
    def nickname(self):
        """
        Nickname of the WBEM server.
        """
        return self._nickname

    @property
    def description(self):
        """
        Short description of the WBEM server.
        """
        return self._description

    @property
    def url(self):
        """
        URL of the WBEM server.

        For details see url parameter of pywbem.WBEMConnections().
        """
        return self._url

    @property
    def user(self):
        """
        User for logging on to the WBEM server.
        """
        return self._user

    @property
    def password(self):
        """
        Password of that user.
        """
        return self._password

    @property
    def cert_file(self):
        """
        Path name of file containing X.509 client certificate to be presented
        to server, or `None` for not presenting a client certificate to the
        server.
        """
        return self._cert_file

    @property
    def key_file(self):
        """
        Path name of file containing X.509 private key of the client
        certificate, or `None` for not presenting a client certificate to the
        server.
        """
        return self._key_file

    @property
    def ca_certs(self):
        """
        Path name of CA certificate file or path name of directory containing
        CA certificate files to be used for verifying the returned server
        certificate, or `None` for using the pywbem default locations.
        """
        return self._ca_certs

    @property
    def no_verification(self):
        """
        Boolean controlling whether the returned server certificate is to be
        verified.
        """
        return self._no_verification

    @property
    def implementation_namespace(self):
        """
        Boolean containing the default namespace or None if no default
        namespace was defined..
        """
        return self._implementation_namespace



#!/usr/bin/env python
#

"""
Unittest functions in _cim_operations.
"""

from __future__ import print_function, absolute_import

import os
import re
import pytest

from ...utils import skip_if_moftab_regenerated
from ..utils.dmtf_mof_schema_def import install_test_dmtf_schema
from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMConnection, ParseError, DEFAULT_NAMESPACE, \
    CIMError, CIMInstanceName, CIMClassName, CIMInstance, \
    CIMClass  # noqa: E402
from pywbem._recorder import LogOperationRecorder  # noqa: E402
from pywbem._recorder import TestClientRecorder as \
    MyTestClientRecorder  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
from pywbem_mock import FakedWBEMConnection  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

OK = True     # mark tests OK when they execute correctly
RUN = True    # Mark OK = False and current test case being created RUN
FAIL = False  # Any test currently FAILING or not tested yet
SKIP = False


TESTCASES_INIT_WBEMCONNECTION = [

    # Test WBEMConnection itit
    # the testcase items are tuples with these items:
    # * desc: Testcase description
    # * files: empty files to be created before the test
    # * init_kwargs: dict of init arguments for WBEMConnection, in addition to
    #   the utl argument
    # * exp_attrs: dict of expected WBEMConnection attributes after init
    # * exp_exc: Type of expected exception, or None
    # * exp_exc_regex: Regexp for the expected exception message, or None

    (
        "No optional init parameters, test defaults",
        [],
        dict(),
        dict(
            creds=None,
            default_namespace='root/cimv2',
            x509=None,
            ca_certs=None,
            no_verification=False,
            timeout=None,
            use_pull_operations=False,
            stats_enabled=False,
            proxies=None,
        ),
        None, None
    ),
    (
        "All optional init parameters, except x509 and ca_certs",
        [],
        dict(
            creds=('myuser', 'mypw'),
            default_namespace='root/myns',
            no_verification=True,
            timeout=30,
            use_pull_operations=True,
            stats_enabled=True,
            proxies=None,
        ),
        dict(
            creds=('myuser', 'mypw'),
            default_namespace='root/myns',
            no_verification=True,
            timeout=30,
            use_pull_operations=True,
            stats_enabled=True,
            proxies=None,
        ),
        None, None
    ),
    (
        "x509 parameter that is a dict with existing cert_file and "
        "existing key_file",
        ['mycertfile.tmp', 'mykeyfile.tmp'],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file='mykeyfile.tmp',
            ),
        ),
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file='mykeyfile.tmp',
            ),
        ),
        None, None
    ),
    (
        "x509 parameter that is a dict with existing cert_file and "
        "omitted key_file",
        ['mycertfile.tmp'],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
            ),
        ),
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
            ),
        ),
        None, None
    ),
    (
        "x509 parameter that is a dict with non-existing cert_file "
        "(invalid) and existing key_file",
        ['mykeyfile.tmp'],
        dict(
            x509=dict(
                cert_file='nonexistingcertfile.tmp',
                key_file='mykeyfile.tmp',
            ),
        ),
        dict(),
        IOError, "Client certificate file .* not found"
    ),
    (
        "x509 parameter that is a dict with existing cert_file "
        "and non-existing key_file (invalid)",
        ['mycertfile.tmp'],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file='nonexistingkeyfile.tmp',
            ),
        ),
        dict(),
        IOError, "Client key file .* not found"
    ),
    (
        "x509 parameter that is a dict with cert_file=None (invalid) "
        "and key_file=None",
        [],
        dict(
            x509=dict(
                cert_file=None,
                key_file=None,
            ),
        ),
        dict(),
        TypeError, "cert_file.* must be a string"
    ),
    (
        "x509 parameter that is a dict with cert_file=3 (invalid) "
        " Must be a string",
        [],
        dict(
            x509=dict(
                cert_file=3,
                key_file=None,
            ),
        ),
        dict(),
        TypeError, "The 'cert_file' item in the x509 parameter must be a string"
    ),
    (
        "x509 parameter that is a dict with cert_file='mycertfile.tmp' "
        " keyfile invalid type",
        [],
        dict(
            x509=dict(
                cert_file='mycertfile.tmp',
                key_file=3,
            ),
        ),
        dict(),
        TypeError, "The 'key_file' item in the x509 parameter must be a string",
    ),
    (
        "x509 parameter that is a dict with no cert_file key "
        "and key_file=None",
        [],
        dict(
            x509=dict(
                key_file=None,
            ),
        ),
        dict(),
        ValueError, "The x509 parameter does not have the required key "
        "'cert_file'"
    ),

    (
        "x509 parameter that is an existing file (invalid)",
        ['mycertfile.tmp'],
        dict(
            x509='mycertfile.tmp',
        ),
        dict(),
        TypeError, "x509 .* must be a dictionary"
    ),
    (
        "ca_certs parameter that is an existing file",
        ['mycacertfile.tmp'],
        dict(
            ca_certs='mycacertfile.tmp',
        ),
        dict(
            ca_certs='mycacertfile.tmp',
        ),
        None, None
    ),
    (
        "ca_certs parameter that is an integer (invalid)",
        [],
        dict(
            ca_certs=42,
        ),
        dict(),
        TypeError, "ca_certs .* invalid type"
    ),
    (
        "ca_certs parameter that is a non-existing file (invalid)",
        [],
        dict(
            ca_certs='mycacertfile.tmp',
        ),
        dict(),
        IOError, "file or directory not found"
    ),
    (
        "proxies parameter that is valid",
        [],
        dict(
            proxies={
                'http': 'http://user:pass@10.10.1.10:3128',
                'https': 'http://user:pass@10.10.1.10:1080',
            },
        ),
        dict(
            proxies={
                'http': 'http://user:pass@10.10.1.10:3128',
                'https': 'http://user:pass@10.10.1.10:1080',
            },
        ),
        None, None
    ),
    (
        "proxies parameter that is an invalid type",
        [],
        dict(
            proxies=42,
        ),
        dict(),
        TypeError, "proxies .* must be a dictionary .*"
    ),
]


class TestCreateConnection(object):
    """
    Test construction of WBEMConnection and those functions that do not
    depend on actually creating a connection
    """

    @pytest.mark.parametrize(
        'desc, files, init_kwargs, exp_attrs, exp_exc, exp_exc_regex',
        TESTCASES_INIT_WBEMCONNECTION)
    def test_init(self, desc, files, init_kwargs, exp_attrs, exp_exc,
                  exp_exc_regex):
        # pylint: disable=no-self-use,unused-argument
        """
        Test initialization of a WBEMConnection object.
        """

        try:
            for fname in files:
                open(fname, 'a').close()  # create empty file

            if exp_exc is not None:
                with pytest.raises(exp_exc) as exc_info:

                    # The code to be tested
                    WBEMConnection('http://localhost', **init_kwargs)

                exc = exc_info.value
                exc_msg = str(exc)
                if exp_exc_regex:
                    assert re.search(exp_exc_regex, exc_msg)
            else:

                # The code to be tested
                conn = WBEMConnection('http://localhost', **init_kwargs)

                for name in exp_attrs:
                    exp_value = exp_attrs[name]
                    act_value = getattr(conn, name)
                    assert act_value == exp_value

        finally:
            for fname in files:
                os.remove(fname)

    @pytest.mark.parametrize(
        'attr_name, value', [
            ('url', 'http://myserver:5988'),
            ('creds', ('x', 'y')),
            ('x509', dict(cert_file='c', key_file='k')),
            ('ca_certs', 'xxx'),
            ('no_verification', True),
            ('timeout', 30),
            ('last_raw_request', '<CIM/>'),
            ('last_raw_reply', '<CIM/>'),
            ('last_raw_reply', '<CIM/>'),
            ('last_request', '<CIM/>'),
            ('last_request_len', 7),
            ('last_reply_len', 7),
        ])
    def test_attrs_readonly(self, attr_name, value):
        # pylint: disable=no-self-use
        """
        Test that certain attributes that were previously settable are now
        read-only and that the original value has not changed when trying to
        modify them.
        """
        conn = WBEMConnection('http://localhost')
        value1 = getattr(conn, attr_name)
        with pytest.raises(AttributeError):

            # The code to be tested
            setattr(conn, attr_name, value)

        value2 = getattr(conn, attr_name)
        assert value2 == value1

    @pytest.mark.parametrize(
        'kwargs, exp_default_namespace', [
            (dict(), DEFAULT_NAMESPACE),
            (dict(default_namespace=None), DEFAULT_NAMESPACE),
            (dict(default_namespace=DEFAULT_NAMESPACE), DEFAULT_NAMESPACE),
            (dict(default_namespace='blah'), 'blah'),
        ])
    def test_init_default_namespace(self, kwargs, exp_default_namespace):
        # pylint: disable=no-self-use
        """Test creation of wbem connection with default_namespace values"""
        conn = WBEMConnection('http://localhost', ('name', 'pw'), **kwargs)
        assert conn.default_namespace == exp_default_namespace

    def test_namespace_slashes_init(self):  # pylint: disable=no-self-use
        """Test stripping of leading and trailing slashes in default namespace
        of wbem connection when initializing"""
        conn = WBEMConnection('http://localhost', None,
                              default_namespace='//root/blah//')
        assert conn.default_namespace == 'root/blah'

    def test_namespace_slashes_set(self):  # pylint: disable=no-self-use
        """Test stripping of leading and trailing slashes in default namespace
        of wbem connection when setting the attribute"""
        conn = WBEMConnection('http://localhost', None,
                              default_namespace=None)
        conn.default_namespace = '//root/blah//'
        assert conn.default_namespace == 'root/blah'

    def test_add_operation_recorder(self):  # pylint: disable=no-self-use
        """Test addition of an operation recorder"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        # pylint: disable=protected-access
        assert len(conn._operation_recorders) == 1

    def test_add_operation_recorders(self):  # pylint: disable=no-self-use
        """Test addition of multiple operation recorders"""
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        tcr_fn = 'blah.yaml'
        tcr_file = MyTestClientRecorder.open_file(tcr_fn, 'a')
        conn.add_operation_recorder(MyTestClientRecorder(tcr_file))
        # pylint: disable=protected-access
        assert len(conn._operation_recorders) == 2
        tcr_file.close()
        os.remove(tcr_fn)

    def test_add_same_twice(self):  # pylint: disable=no-self-use
        """
        Test addition of same recorder twice. This not allowed so
        generates exception
        """
        conn = WBEMConnection('http://localhost')
        conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))
        # pylint: disable=protected-access
        assert len(conn._operation_recorders) == 1

        with pytest.raises(ValueError):
            conn.add_operation_recorder(LogOperationRecorder('fake_conn_id'))


class TestGetRsltParams(object):
    """Test WBEMConnection._get_rslt_params method."""

    @pytest.mark.parametrize(
        'irval', [None, 'bla', [], {}])
    @pytest.mark.parametrize(
        # enumctxt input, eos input, expected eos, expected exception type
        'ec, eos, eos_exp, exctype_exp', [
            # following are successul returns
            [u'contextblah', u'False', False, None],  # normal enumctxt rtn
            ["", u'True', True, None],                # eos true, enmctxt empty
            [None, u'True', True, None],              # eos true, enmctxt None
            [u'contextblah', u'True', True, None],    # eos tru, cts with value
            # following are exceptions
            [None, None, None, ParseError],      # no values in eos or enumctxt
            [None, u'False', None, ParseError],  # no value in ec and eos False
        ]
    )
    def test_with_params(self, irval, ec, eos, eos_exp, exctype_exp):
        # pylint: disable=no-self-use,protected-access
        """
        Test combminations of IRETURNVALUE, EOS and EnumerationContext for
        both good and fail responses.
        """
        conn = WBEMConnection('http://localhost')
        result = [
            (u'IRETURNVALUE', {}, irval),
            (u'EnumerationContext', None, ec),
            (u'EndOfSequence', None, eos)
        ]

        if exctype_exp:
            with pytest.raises(exctype_exp):

                # pylint: disable=protected-access
                result = conn._get_rslt_params(result, 'root/blah')

        else:

            # pylint: disable=protected-access
            result = conn._get_rslt_params(result, 'root/blah')

            # the _get_rslt_params method sets context to None if eos True
            ecs = None if eos_exp is True else (ec, 'root/blah')
            assert result == (irval, eos_exp, ecs)

    def test_missing(self):
        # pylint: disable=no-self-use
        """
        Test both Enum context and EndOfSequence completely missing.
        Generates exception
        """
        conn = WBEMConnection('http://localhost')
        result = [
            (u'IRETURNVALUE', {}, {})
        ]

        with pytest.raises(ParseError):

            # pylint: disable=protected-access
            result = conn._get_rslt_params(result, 'namespace')


def build_repo():
    """Fixture to initialize the mock environment and install classes.
       Definde as a function without the pytest.fixture decorator becasue
       there is no way to include a fixture in the signature of a test
       function that uses simplified_test_function. Simplified_test_function
       reports a parameter count difference because of the extram parameter
       that is the fixture and fixtures may not be included in the testcases.
    """
    schema = install_test_dmtf_schema()
    namespace = "root/cimv2"
    partial_schema = """
        #pragma locale ("en_US")
        #pragma include ("Interop/CIM_ObjectManager.mof")
        #pragma include ("Interop/CIM_RegisteredProfile.mof")
        """

    conn = FakedWBEMConnection(default_namespace=namespace)
    conn.compile_mof_string(partial_schema, namespace=namespace,
                            search_paths=[schema.schema_mof_dir])
    return conn


TESTCASES_IS_SUBCLASS = [

    # Testcases for WBEMConnection.is_subclass()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to is_subclass().
    #   * init_kwargs: Dict of keyword arguments to is_subclass().
    #   * exp_attrs: Dict of expected attributes of resulting object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger

    (
        "Test of valid call with strings that returns True",
        dict(
            init_args=['CIM_ObjectManager', 'CIM_ManagedElement'],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call with CIM objects that returns True",
        dict(
            init_args=[
                CIMClass('CIM_ObjectManager', superclass='CIM_WBEMService'),
                CIMClass('CIM_ManagedElement', superclass=None),
            ],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call with CIM object and string that returns True",
        dict(
            init_args=[
                CIMClass('CIM_ObjectManager', superclass='CIM_WBEMService'),
                'CIM_ManagedElement',
            ],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call with string and CIM object that returns True",
        dict(
            init_args=[
                'CIM_ObjectManager',
                CIMClass('CIM_ManagedElement', superclass=None),
            ],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of valid call that returns False",
        dict(
            init_args=['CIM_ManagedElement', 'CIM_ObjectManager'],
            exp_attrs=False,
        ),
        None, None, True
    ),
    (
        "Test of class equal to superclass",
        dict(
            init_args=['CIM_ObjectManager', 'CIM_ObjectManager'],
            exp_attrs=True,
        ),
        None, None, True
    ),
    (
        "Test of superclass not in repo, returns False",
        dict(
            init_args=['CIM_ManagedElement', 'CIM_Blah'],
            exp_attrs=False,
        ),
        None, None, True
    ),
    (
        "Test of class not in repo (error)",
        dict(
            init_args=['CIM_Blah', 'CIM_ManagedElement'],
            exp_attrs=None,
        ),
        CIMError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_IS_SUBCLASS)
@simplified_test_function
def test_is_subclass(testcase, init_args, exp_attrs):
    # pylint: disable=unused-argument
    """
    Test the method WBEMConnection.is_subclass().
    """

    skip_if_moftab_regenerated()

    conn = build_repo()   # we wanted this to be fixture.
    # The code to be tested
    result = conn.is_subclass(conn.default_namespace, *init_args)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None
    assert exp_attrs == result


# TODO List of failure tests to be tested with following test
# methodcall. test objectname
# Test Iter invalid maxcount >- 0 and OperationTimeout < 0
# line 2008, methodcall request with classname no namespace  - not a failue case
# line 2010, methodcall request with classname, use string  - not a failure case
# request with object name not CIMInstanceName, CIMClassName, not string - DONE
# test _iparam_objectname type error if not CIMClassname, instancename,
#     string - DONE
# iparam_classname not CIMClassName or string - DONE
# iparam_instancename not CIMInstanceName or None  - DONE
# modified instance. no path,
# Modified instance no class in path
# modified_instance no classname in modifiedinstance line 3022
# line 4595, iterenumerateinstances CIMClassName input no namespace
# line 4612, close iterenum early.  Not sure this can be done.

TESTCASES_REQUEST_INVALID_PARAMS = [

    # Testcases for WBEMConnection  method calls with invalid params.

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_kwargs: keyword args for WBEMConnection __init
    #   * method: operation method to call
    #   * args=: Arguments for the operation method being called
    #   * kwargs: Dict of keyword arguments for the WBEMConnection __init
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for
    #   debugger
    (
        "Test invalid invokemethod objectname (integer)",
        dict(
            init_kwargs={},
            method='invokemethod',
            args=["MethodName", 1],
            kwargs={},
        ),
        TypeError, None, OK
    ),
    (
        "Test invalid invokemethod objectname is None",
        dict(
            init_kwargs={},
            method='invokemethod',
            args=["MethodName", None],
            kwargs={},
        ),
        TypeError, None, OK
    ),
    (
        "Test invalid invokemethod objectname None",
        dict(
            init_kwargs={},
            method='invokemethod',
            args=[None, "ObjectName"],
            kwargs={},
        ),
        TypeError, None, False  # Fails with connection failure.
    ),
    (
        "Test enumerateinstances, fails, classname is not CIMClassname or Str",
        dict(
            init_kwargs={},
            method='enumerateinstances',
            args=[CIMInstanceName("CIM_Blah")],
            kwargs={},
        ),
        TypeError, None, OK
    ),
    (
        "Test GetInstance, invalid InstanceName tyhpe",
        dict(
            init_kwargs={},
            method='getinstance',
            args=[CIMClassName("CIM_Blah")],
            kwargs={},
        ),
        TypeError, None, OK
    ),

    (
        "Test ModifyInstance, No path in ModifiedInstance",
        dict(
            init_kwargs={},
            method='modifyinstance',
            args=[CIMInstance("CIMBlah")],
            kwargs={},
        ),
        ValueError, None, OK
    ),

    (
        "Test EnumerateInstances, Invalid args type",
        dict(
            init_kwargs={},
            method='enumerateinstances',
            args=[CIMClass("CIMBlah")],
            kwargs={},
        ),
        TypeError, None, OK
    ),

    (
        "Test pullinstancewithpath, Invalid context type",
        dict(
            init_kwargs={},
            method='pullinstanceswithpath',
            args=[1, 0],
            kwargs={},
        ),
        TypeError, None, OK
    ),

    (
        "Test pullinstancewithpath, context list len ",
        dict(
            init_kwargs={},
            method='pullinstanceswithpath',
            args=[[], 0],
            kwargs={},
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, invalid timeout.lt 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(OperationTimeout=-1),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, invalid maxobjectcount == 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(MaxObjectCount=0),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, Pull not allowed. with Filterquery",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQuery="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, Pull not allowed. with FilterQueryLang",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQueryLanguage="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, invalid timeout.lt 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(OperationTimeout=-1),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, invalid maxobjectcount == 0",
        dict(
            init_kwargs={},
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(MaxObjectCount=0),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, Pull not allowed. with Filterquery",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQuery="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, Pull not allowed. with "
        "FilterQueryLang",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(FilterQueryLanguage="blah"),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstances, Pull not allowed. with "
        "ContinueOnError True",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstances',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(ContinueOnError=True),
        ),
        ValueError, None, OK
    ),

    (
        "Test IterEnumerateInstancePaths, Pull not allowed. with "
        "ContinueOnError True",
        dict(
            init_kwargs=dict(use_pull_operations=False),
            method='iterenumerateinstancepaths',
            args=[CIMClassName("CIMBlah")],
            kwargs=dict(ContinueOnError=True),
        ),
        ValueError, None, OK
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_REQUEST_INVALID_PARAMS)
@simplified_test_function
def test_request_invalid_params(testcase, init_kwargs, method, args, kwargs):
    """Execute the defined method expecting exception"""

    conn = WBEMConnection('http://localhost', **init_kwargs)

    if method == 'invokemethod':
        _ = conn.InvokeMethod(*args, **kwargs)

    elif method == 'enumerateinstances':
        _ = conn.EnumerateInstances(*args, **kwargs)

    elif method == 'getinstance':
        _ = conn.GetInstance(*args, **kwargs)

    elif method == 'modifyinstance':
        conn.ModifyInstance(*args, **kwargs)

    elif method == 'openenumerateinstances':
        _ = conn.OpenEnumerateInstances(*args, **kwargs)

    elif method == 'pullinstanceswithpath':
        _ = conn.PullInstancesWithPath(*args, **kwargs)

    elif method == 'iterenumerateinstances':
        _ = list(conn.IterEnumerateInstances(*args, **kwargs))

    elif method == 'iterenumerateinstancepaths':
        _ = list(conn.IterEnumerateInstancePaths(*args, **kwargs))

    else:
        assert False, "Test failed. Incomplete. "  \
            "method {0} not found".format(method)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# There are submodules, but clients shouldn't need to know about them.
# Importing just this module is enough.
# These are explicitly safe for 'import *'

"""
Version of the pywbem package.
"""

#: The full version of this package including any development levels, as a
#: :term:`string`.
#:
#: Possible formats for this version string are:
#:
#: * "M.N.P.devD": Development level D of a not yet released version M.N.P
#: * "M.N.P": A released version M.N.P
__version__ = '1.1.1'



#
# (C) Copyright 2003-2020 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
pywbem_mock - Unit test support for users of pywbem package that mocks
the implementation of the WBEMConnection calls to WBEM Servers.
"""

from __future__ import absolute_import

from ._wbemconnection_mock import *        # noqa: F403,F401
from ._dmtf_cim_schema import *            # noqa: F403,F401
from ._resolvermixin import *              # noqa: F403,F401
from ._mockmofwbemconnection import *      # noqa: F403,F401
from ._baserepository import *             # noqa: F403,F401
from ._inmemoryrepository import *         # noqa: F403,F401
from ._baseprovider import *               # noqa: F403,F401
from ._mainprovider import *               # noqa: F403,F401
from ._providerdispatcher import *         # noqa: F403,F401
from ._providerregistry import *           # noqa: F403,F401
from ._providerdependentregistry import *  # noqa: F403,F401
from ._instancewriteprovider import *      # noqa: F403,F401
from ._methodprovider import *             # noqa: F403,F401
from ._namespaceprovider import *          # noqa: F403,F401
from ._utils import *                      # noqa: F403,F401



"""
Test memory leaks for cim_xml.py module.
"""

from __future__ import absolute_import, print_function

import pytest
import yagot

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils import import_installed
pywbem = import_installed('pywbem')
from pywbem._cim_xml import CIMElement, VALUE  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


@yagot.garbage_checked()
def test_leaks_CIMElement_empty():
    """
    Test function with an empty CIMElement object.
    """
    _ = CIMElement('dummy')


@yagot.garbage_checked()
def test_leaks_VALUE_empty():
    """
    Test function with a VALUE object that is empty (ie. no text child).
    """
    _ = VALUE(pcdata=None)


@pytest.mark.skip("Temporarily skipped")
@yagot.garbage_checked()
def test_leaks_VALUE_string():
    """
    Test function with a VALUE object of string.
    """
    _ = VALUE(pcdata='abc')



"""
    Define the classes and instances to mock a WEMServer as the basis for
    other test mocks
"""

from __future__ import print_function, absolute_import

import os

from .dmtf_mof_schema_def import DMTF_TEST_SCHEMA_VER, DMTFCIMSchema

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMServer, ValueMapping, CIMInstance, CIMQualifier, \
    CIMInstanceName  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
from pywbem_mock import FakedWBEMConnection  # noqa: E402
from pywbem_mock import OBJECTMANAGERCREATIONCLASSNAME, \
    SYSTEMCREATIONCLASSNAME, OBJECTMANAGERNAME, \
    SYSTEMNAME  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


# Location of DMTF schema directory used by all tests.
# This directory is permanent and should not be removed.
TESTSUITE_SCHEMA_DIR = os.path.join('tests', 'schema')

# The following dictionary represents the data required to build a
# set of classes and instances for a mock WbemServer. This dictionary defines
# the following elements of a wbem server:
#
#  dmtf_schema: The DMTF schema version (ex (2, 49, 0) of the schema that
#  will be installed and the directory into which it will be installed
#
#  system_name: The name of the system (used by the CIM_ObjectManager)
#
#  object_manager: A dictionary that defines variable element for the
#  CIM_ObjectManager class.
#
#  interop_namespace: The interop namespace. Note that that if the interop
#  namespace is defined in the WbemServerMock init method that overrides any
#  value in this dictionary
#
#  other_namespaces: Any other namespaces that the users wants to be defined
#  in the CIM repository
#
#  registered_profiles: The Organization, profile name, profile version for any
#  registered profiles
#
#  referenced_profiles: Definition of CIM_ReferencedProfile associations
#  between CIM_RegisteredProfiles. This is used to test get_central_instances
#
#
DEFAULT_WBEM_SERVER_MOCK_DICT = {
    'dmtf_schema': {'version': DMTF_TEST_SCHEMA_VER,
                    'dir': TESTSUITE_SCHEMA_DIR},
    'tst_schema': {
        'dir': os.path.join(TESTSUITE_SCHEMA_DIR, 'FakeWBEMServer'),
        'files': []},

    'class_names': ['CIM_Namespace',
                    'CIM_ObjectManager',
                    'CIM_RegisteredProfile',
                    'CIM_ElementConformsToProfile',
                    'CIM_ReferencedProfile',
                    'CIM_ComputerSystem'],
    'class-mof': ["class XXX_StorageComputerSystem : CIM_ComputerSystem{};", ],
    'system_name': SYSTEMNAME,
    'object_manager': {'Name': OBJECTMANAGERNAME,
                       'ElementName': 'Mock_Test',
                       'Description': 'Mock_Test CIM Server Version 2.15.0'
                                      ' Released', },
    'interop_namspace': 'interop',
    'other_namespaces': [],
    'registered_profiles': [('DMTF', 'Indications', '1.1.0'),
                            ('DMTF', 'Profile Registration', '1.0.0'),
                            ('SNIA', 'Server', '1.2.0'),
                            ('SNIA', 'Server', '1.1.0'),
                            ('SNIA', 'SMI-S', '1.2.0'),
                            ('SNIA', 'Array', '1.4.0'),
                            ('SNIA', 'Software', '1.4.0'),
                            ('DMTF', 'Component', '1.4.0'), ],

    'referenced_profiles': [
        (('SNIA', 'Server', '1.2.0'), ('DMTF', 'Indications', '1.1.0')),
        (('SNIA', 'Server', '1.2.0'), ('SNIA', 'Array', '1.4.0')),
        (('SNIA', 'Array', '1.4.0'), ('DMTF', 'Component', '1.4.0')),
        (('SNIA', 'Array', '1.4.0'), ('SNIA', 'Software', '1.4.0')),

    ],
    # define profile relation to central class instance
    # First element is a profile, second is name and keybindings of a
    # CIMInstanceName
    'element_conforms_to_profile': [
        (('SNIA', 'Server', '1.2.0'),
         ("XXX_StorageComputerSystem", {'Name': "10.1.2.3",
                                        'CreationClassName':
                                            "XXX_StorageComputerSystem"})), ],
    # List of CIMInstances. Each entry is a CIM instance with classname,
    # and properties. All properties required to build the path must be
    # defined. No other properties are required for this test.
    # TODO: We may expand this for more scoping tests.
    'central-instances': [
        CIMInstance(
            'XXX_StorageComputerSystem',
            properties={
                'Name': "10.1.2.3",
                'CreationClassName': "XXX_StorageComputerSystem",
                'NameFormat': "IP"}),
    ],
    'scoping-instances': []
}


class WbemServerMock(object):
    """
    Class that mocks the classes and methods used by the pywbem
    WBEMServer class so that the WBEMServer class will produce valid data
    for the server CIM_ObjectManager, CIM_Namespace, CIM_RegisteredProfile
    instances.

    This can be used to test the WbemServer class but is also required for
    other tests such as the Subscription manager since that class is based
    on getting data from the WbemServer class (ex. namespace)

    It allows building a the instance data for a particular server either
    from user defined input or from standard data predefined for pywbem
    tests
    """

    def __init__(self, interop_ns=None, server_mock_data=None):
        """
        Build the class repository for with the classes defined for
        the WBEMSERVER.  This is built either from a dictionary of data
        that represents the mock wbem server with the elements defined in
        DEFAULT_WBEM_SERVER_MOCK_DICT or if server_mock_data is None from
        the DEFAULT_WBEM_SERVER_MOCK_DICT dictionary.
        """
        if server_mock_data is None:
            self.server_mock_data = DEFAULT_WBEM_SERVER_MOCK_DICT
        else:
            self.server_mock_data = server_mock_data

        self.system_name = self.server_mock_data['system_name']
        self.object_manager_name = \
            self.server_mock_data['object_manager']['Name']

        # override dictionary interop namespace with init parameter
        if interop_ns:
            self.interop_ns = interop_ns
        else:
            self.interop_ns = self.server_mock_data['interop_ns']

        self.dmtf_schema_ver = self.server_mock_data['dmtf_schema']['version']
        self.schema_dir = self.server_mock_data['dmtf_schema']['dir']
        self.tst_schema_dir = self.server_mock_data['tst_schema']['dir']
        self.tst_schema_files = self.server_mock_data['tst_schema']['files']
        self.registered_profiles = self.server_mock_data['registered_profiles']
        self.wbem_server = self.build_mock()

    def __str__(self):
        ret_str = 'object_manager_name=%r, interop_ns=%r, system_name=%r, ' \
            'dmtf_schema_ver=%r, schema_dir=%r, wbem_server=%s' % \
            (self.object_manager_name, self.interop_ns, self.system_name,
             self.dmtf_schema_ver, self.schema_dir,
             getattr(self, 'wbem_server', None))
        return ret_str

    def __repr__(self):
        """
        Return a representation of the class object
        with all attributes, that is suitable for debugging.
        """
        ret_str = 'WBEMServerMock(object_manager_name=%r, interop_ns=%r, ' \
            'system_name=%r, dmtf_schema_ver=%r, schema_dir=%r, ' \
            'wbem_server=%r, registered_profiles=%r)' % \
            (self.object_manager_name, self.interop_ns, self.system_name,
             self.dmtf_schema_ver, self.schema_dir,
             getattr(self, 'wbem_server', None), self.registered_profiles)
        return ret_str

    def build_class_repo(self, default_namespace):
        """
        Build the schema qualifier and class objects in the repository
        from a DMTF schema.
        This requires only that the leaf objects be defined in a mof
        include file since the compiler finds the files for qualifiers
        and dependent classes.

        Returns:
            Instance of FakedWBEMConnection object.
        """
        # pylint: disable=protected-access

        FakedWBEMConnection._reset_logging_config()
        conn = FakedWBEMConnection(default_namespace=default_namespace)

        schema = DMTFCIMSchema(self.dmtf_schema_ver, self.schema_dir,
                               verbose=False)

        conn.compile_schema_classes(
            self.server_mock_data['class_names'],
            schema.schema_pragma_file,
            verbose=False)

        for fn in self.tst_schema_files:
            pg_file = os.path.join(self.tst_schema_dir, fn)
            conn.compile_mof_file(pg_file, namespace=default_namespace,
                                  search_paths=[self.tst_schema_dir],
                                  verbose=False)

        # compile the mof defined in the 'class-mof definitions
        for mof in self.server_mock_data['class-mof']:
            conn.compile_mof_string(mof, namespace=default_namespace,
                                    verbose=False)

        return conn

    @staticmethod
    def inst_from_classname(conn, class_name, namespace=None,
                            property_list=None,
                            property_values=None,
                            include_missing_properties=True,
                            include_path=True):
        """
        Build instance from classname using class_name property to get class
        from a repository.
        """
        cls = conn.GetClass(class_name, namespace=namespace, LocalOnly=False,
                            IncludeQualifiers=True, IncludeClassOrigin=True,
                            PropertyList=property_list)

        return CIMInstance.from_class(
            cls, namespace=namespace, property_values=property_values,
            include_missing_properties=include_missing_properties,
            include_path=include_path)

    def build_obj_mgr_inst(self, conn, system_name, object_manager_name,
                           object_manager_element_name,
                           object_manager_description):
        """
        Build a CIMObjectManager instance for the mock wbem server using
        fixed data defined in this method and data from the init parameter
        mock data.
        """
        omdict = {"SystemCreationClassName": SYSTEMCREATIONCLASSNAME,
                  "CreationClassName": OBJECTMANAGERCREATIONCLASSNAME,
                  "SystemName": system_name,
                  "Name": object_manager_name,
                  "ElementName": object_manager_element_name,
                  "Description": object_manager_description}

        ominst = self.inst_from_classname(conn, "CIM_ObjectManager",
                                          namespace=self.interop_ns,
                                          property_values=omdict,
                                          include_missing_properties=False,
                                          include_path=True)

        conn.add_cimobjects(ominst, namespace=self.interop_ns)

        rtn_ominsts = conn.EnumerateInstances("CIM_ObjectManager",
                                              namespace=self.interop_ns)
        assert len(rtn_ominsts) == 1, \
            "Expected 1 ObjetManager instance, got %r" % rtn_ominsts

        return ominst

    def build_reg_profile_insts(self, conn, profiles):
        """
        Build and install in repository the registered profiles define by
        the profiles parameter. A dictionary of tuples where each tuple
        contains RegisteredOrganization, RegisteredName, RegisteredVersion

        Parameters:
          conn:
          profiles (dict of lists where each list contains org, name, version
             for a profiles)
        """
        # Map ValueMap to Value
        org_vm = ValueMapping.for_property(conn, self.interop_ns,
                                           'CIM_RegisteredProfile',
                                           'RegisteredOrganization')

        # This is a workaround hack to get ValueMap from Value
        org_vm_dict = {}   # reverse mapping dictionary (valueMap from Value)
        for value in range(0, 22):
            org_vm_dict[org_vm.tovalues(value)] = value

        for p in profiles:
            instance_id = '%s+%s+%s' % (p[0], p[1], p[2])
            reg_prof_dict = {'RegisteredOrganization': org_vm_dict[p[0]],
                             'RegisteredName': p[1],
                             'RegisteredVersion': p[2],
                             'InstanceID': instance_id}
            rpinst = self.inst_from_classname(conn, "CIM_RegisteredProfile",
                                              namespace=self.interop_ns,
                                              property_values=reg_prof_dict,
                                              include_missing_properties=False,
                                              include_path=True)

            conn.add_cimobjects(rpinst, namespace=self.interop_ns)

        rtn_rpinsts = conn.EnumerateInstances("CIM_RegisteredProfile",
                                              namespace=self.interop_ns)

        assert len(rtn_rpinsts) == len(profiles), \
            "Expected registered profiles: %r, got %s" % (len(profiles),
                                                          len(rtn_rpinsts))

    def build_elementconformstoprofile_inst(self, conn, profile_path,
                                            element_path):
        """
        Build an instance of CIM_ElementConformsToProfile and insert into
        repository
        """
        class_name = 'CIM_ElementConformsToProfile'
        element_conforms_dict = {'ConformantStandard': profile_path,
                                 'ManagedElement': element_path}

        # TODO modify this when issue #1540  (resolve qualifiers)fixed
        # inst = self.inst_from_classname(conn, class_name,
        #                                namespace=self.interop_ns,
        #                                property_values=element_conforms_dict,
        #                                include_missing_properties=False,
        #                                include_path=True)
        cls = conn.GetClass(class_name, namespace=self.interop_ns,
                            LocalOnly=False, IncludeQualifiers=True,
                            IncludeClassOrigin=True)

        for pvalue in cls.properties.values():
            if pvalue.type == 'reference':
                if "key" not in pvalue.qualifiers:
                    pvalue.qualifiers['Key'] = \
                        CIMQualifier('Key', True, propagated=True)

        inst = CIMInstance.from_class(
            cls, namespace=self.interop_ns,
            property_values=element_conforms_dict,
            include_missing_properties=False,
            include_path=True)
        # TODO end of temp code

        conn.add_cimobjects(inst, namespace=self.interop_ns)

        assert conn.EnumerateInstances(class_name, namespace=self.interop_ns)
        assert conn.GetInstance(inst.path)

    def build_referenced_profile_insts(self, server, referenced_profiles):
        """
        Build and install in repository the referemced profile instances
        defined by the referemces parameter. A dictionary of tuples where each
        tuple contains Antecedent and Dependent reference in terms of the
        profile name as a tuple (org, name, version).

        Parameters:
          conn:
          profiles (dict of tuples where each tuple defines the antecedent
          and dependent)
        """
        class_name = 'CIM_ReferencedProfile'
        for profile_name in referenced_profiles:
            antecedent = profile_name[0]
            dependent = profile_name[1]
            antecedent_inst = server.get_selected_profiles(
                registered_org=antecedent[0],
                registered_name=antecedent[1],
                registered_version=antecedent[2])
            dependent_inst = server.get_selected_profiles(
                registered_org=dependent[0],
                registered_name=dependent[1],
                registered_version=dependent[2])

            assert len(antecedent_inst) == 1, \
                "Antecedent: {0}".format(antecedent)
            assert len(dependent_inst) == 1, \
                "Dependent: {0}".format(dependent)

            ref_profile_dict = {'Antecedent': antecedent_inst[0].path,
                                'Dependent': dependent_inst[0].path}

            # TODO replace the setting of key qualifier with the commented
            # code with #issue 1540 is fixed, i.e the key qualifier is
            # included in the class.
            # inst = self.inst_from_classname(server.conn, class_name,
            #                                namespace=self.interop_ns,
            #                                property_values=ref_profile_dict,
            #                                include_missing_properties=False,
            #                                include_path=True)

            cls = server.conn.GetClass(class_name, namespace=self.interop_ns,
                                       LocalOnly=False, IncludeQualifiers=True,
                                       IncludeClassOrigin=True,
                                       PropertyList=None)

            for pvalue in cls.properties.values():
                if pvalue.type == 'reference':
                    if "key" not in pvalue.qualifiers:
                        pvalue.qualifiers['Key'] = \
                            CIMQualifier('Key', True, propagated=True)

            inst = CIMInstance.from_class(
                cls, namespace=self.interop_ns,
                property_values=ref_profile_dict,
                include_missing_properties=False,
                include_path=True)
            # TODO end of code to drop for #1540 fix

            server.conn.add_cimobjects(inst, namespace=self.interop_ns)

            assert server.conn.EnumerateInstances(class_name,
                                                  namespace=self.interop_ns)
            assert server.conn.GetInstance(inst.path)

    def build_central_instances(self, conn, central_instances):
        """
        Build the central_instances from the definitions provided in the list
        central_instance where each definition is a python CIMInstance object
        and add them to the repositoryu. This method adds the path to each
        """
        for inst in central_instances:
            cls = conn.GetClass(inst.classname, namespace=self.interop_ns,
                                LocalOnly=False, IncludeQualifiers=True,
                                IncludeClassOrigin=True)
            inst.path = CIMInstanceName.from_instance(
                cls, inst, namespace=self.interop_ns, strict=True)

            conn.add_cimobjects(inst, namespace=self.interop_ns)

    def build_mock(self):
        """
            Builds the classes and instances for a mock WBEMServer from data
            in the init parameter. This calls the builder for:
              the object manager
              the namespaces
              the profiles
        """
        conn = self.build_class_repo(self.interop_ns)
        server = WBEMServer(conn)
        # NOTE: The wbemserver is not complete until the instances for at
        # least object manager and namespaces have been inserted. Any attempt
        # to display the instance object before that will fail because the
        # enumerate namespaces will be inconsistent

        # Build CIM_ObjectManager instance into the interop namespace since
        # this is required to build namespace instances
        om_inst = self.build_obj_mgr_inst(
            conn,
            self.system_name,
            self.object_manager_name,
            self.server_mock_data['object_manager']['ElementName'],
            self.server_mock_data['object_manager']['Description'])

        # build CIM_Namespace instances based on the init parameters
        namespaces = [self.interop_ns]
        if self.server_mock_data['other_namespaces']:
            namespaces.extend(self.server_mock_data['other_namespaces'])

        conn.install_namespace_provider(self.interop_ns,
                                        None,
                                        verbose=None)

        self.build_reg_profile_insts(conn, self.registered_profiles)

        self.build_referenced_profile_insts(
            server, self.server_mock_data['referenced_profiles'])

        self.build_central_instances(
            conn, self.server_mock_data['central-instances'])

        # Element conforms for SNIA server to object manager
        prof_inst = server.get_selected_profiles(registered_org='SNIA',
                                                 registered_name='Server',
                                                 registered_version='1.1.0')
        # TODO this is simplistic form and only builds one instance
        # conforms to.  Should expand but need better way to define instance
        # at other end.
        self.build_elementconformstoprofile_inst(conn,
                                                 prof_inst[0].path,
                                                 om_inst.path)
        for item in self.server_mock_data['element_conforms_to_profile']:
            profile_name = item[0]
            # TODO we are fixing the host name here.  Not good
            central_inst_path = CIMInstanceName(item[1][0],
                                                keybindings=item[1][1],
                                                host='FakedUrl',
                                                namespace=server.interop_ns)
            prof_insts = server.get_selected_profiles(
                registered_org=profile_name[0],
                registered_name=profile_name[1],
                registered_version=profile_name[2])
            assert len(prof_insts) == 1

            self.build_elementconformstoprofile_inst(conn,
                                                     prof_insts[0].path,
                                                     central_inst_path)
        return server



"""
Extensions for unittest module.

The ideas for these extensions are based on functions from the comfychair
module by Martin Pool <mbp@samba.org>.
"""

from __future__ import absolute_import

import re
import os

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem._nocasedict import NocaseDict  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


class RegexpMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for regular expression tests."""

    @staticmethod
    def assert_regexp_contains(text, pattern, msg=None):
        """Assert that a string text *contains* a particular regular expression
        pattern.

        Parameters:
            text         string: string text to be searched
            pattern      string: regular expression
            msg          string: additional message text (normally not needed)

        Raises:
            AssertionError if not matched
          """
        assert re.search(pattern, text), \
            "String text does not contain regexp pattern\n" \
            "    text:    %r\n" \
            "    pattern: %r\n" \
            "%s" % (text, pattern, msg)

    @staticmethod
    def assert_regexp_matches(text, pattern, msg=None):
        """Assert that a string text matches a particular regular expression
        pattern.

        Parameters:
            text         string: string text to be matched
            pattern      string: regular expression
            msg          string: additional message text (normally not needed)

        Raises:
            AssertionError if not matched
          """
        assert re.match(pattern, text), \
            "String text does not match regexp pattern\n" \
            "    text:    %r\n" \
            "    pattern: %r\n" \
            "%s" % (text, pattern, msg)


class FileMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for file related tests."""

    @staticmethod
    def assert_file_exists(filename):
        """Test assert file exists"""
        assert os.path.exists(filename), \
            "File does not exist but should: %s" % filename

    @staticmethod
    def assert_file_not_exists(filename):
        """Test assert file exists but should not"""
        assert not os.path.exists(filename), \
            "File exists but should not: %s" % filename


class CIMObjectMixin(object):
    """
    Mixin class for testing CIM objects.
    """

    @staticmethod
    def assert_CIMInstanceName_attrs(
            obj, classname, keybindings, host=None, namespace=None):
        """
        Verify the attributes of the CIMInstanceName object in `obj` against
        expected values passed as the remaining arguments.
        """
        assert obj.classname == classname
        assert obj.keybindings == keybindings
        assert obj.host == host
        assert obj.namespace == namespace

    @staticmethod
    def assert_CIMProperty_attrs(
            obj, name, value, type_=None, class_origin=None,
            array_size=None, propagated=None, is_array=False,
            reference_class=None, qualifiers=None, embedded_object=None):
        """
        Verify the attributes of the CIMProperty object in `obj` against
        expected values passed as the remaining arguments.
        """
        assert obj.name == name
        assert obj.value == value
        assert obj.type == type_
        assert obj.class_origin == class_origin
        assert obj.array_size == array_size
        assert obj.propagated == propagated
        assert obj.is_array == is_array
        assert obj.reference_class == reference_class
        assert obj.qualifiers == NocaseDict(qualifiers)
        assert obj.embedded_object == embedded_object

    def assert_CIMClass_obj(self, act_class, exp_class):
        """
        Verify that two CIMClass objects are equal.
        """
        assert act_class.classname == exp_class.classname
        context = "class %s" % act_class.classname
        assert act_class.superclass == exp_class.superclass
        self.assert_CIMProperty_dict(
            act_class.properties, exp_class.properties, context)
        self.assert_CIMMethod_dict(
            act_class.methods, exp_class.methods, context)
        self.assert_CIMQualifier_dict(
            act_class.qualifiers, exp_class.qualifiers, context)

    def assert_CIMProperty_dict(self, act_properties, exp_properties, context):
        """
        Verify that two dicts of CIMProperty objects are equal.
        """

        self.assert_dict_keys(act_properties, exp_properties,
                              "%s, property names" % context)
        for propname in act_properties:
            self.assert_CIMProperty_obj(
                act_properties[propname], exp_properties[propname],
                "%s, property %s" % (context, propname))

    def assert_CIMMethod_dict(self, act_methods, exp_methods, context):
        """
        Verify that two dicts of CIMMethod objects are equal.
        """

        self.assert_dict_keys(act_methods, exp_methods,
                              "%s, method names" % context)
        for methname in act_methods:
            self.assert_CIMMethod_obj(
                act_methods[methname], exp_methods[methname],
                "%s, method %s" % (context, methname))

    def assert_CIMParameter_dict(self, act_parameters, exp_parameters, context):
        """
        Verify that two dicts of CIMParameter objects are equal.
        """

        self.assert_dict_keys(act_parameters, exp_parameters,
                              "%s, parameter names" % context)
        for parmname in act_parameters:
            self.assert_CIMParameter_obj(
                act_parameters[parmname], exp_parameters[parmname],
                "%s, parameter %s" % (context, parmname))

    def assert_CIMQualifier_dict(self, act_qualifiers, exp_qualifiers, context):
        """
        Verify that two dicts of CIMQualifier objects are equal.
        """

        self.assert_dict_keys(act_qualifiers, exp_qualifiers,
                              "%s, qualifier names" % context)
        for qualname in act_qualifiers:
            self.assert_CIMQualifier_obj(
                act_qualifiers[qualname], exp_qualifiers[qualname],
                "%s, qualifier %s" % (context, qualname))

    def assert_CIMProperty_obj(self, act_property, exp_property, context):
        """
        Verify that two CIMProperty declaration objects are equal.
        """

        assert act_property.name == exp_property.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_property.name, exp_property.name)
        assert act_property.value == exp_property.value, \
            "%s, value attribute: %s (expected: %s)" % \
            (context, act_property.value, exp_property.value)
        assert act_property.type == exp_property.type, \
            "%s, type attribute: %s (expected: %s)" % \
            (context, act_property.type, exp_property.type)
        assert act_property.reference_class == exp_property.reference_class, \
            "%s, reference_class attribute: %s (expected: %s)" % \
            (context,
             act_property.reference_class, exp_property.reference_class)
        assert act_property.embedded_object == exp_property.embedded_object, \
            "%s, embedded_object attribute: %s (expected: %s)" % \
            (context,
             act_property.embedded_object, exp_property.embedded_object)
        assert act_property.is_array == exp_property.is_array, \
            "%s, is_array attribute: %s (expected: %s)" % \
            (context, act_property.is_array, exp_property.is_array)
        assert act_property.array_size == exp_property.array_size, \
            "%s, array_size attribute: %s (expected: %s)" % \
            (context, act_property.array_size, exp_property.array_size)
        assert act_property.class_origin == exp_property.class_origin, \
            "%s, class_origin attribute: %s (expected: %s)" % \
            (context, act_property.class_origin, exp_property.class_origin)
        assert act_property.propagated == exp_property.propagated, \
            "%s, propagated attribute: %s (expected: %s)" % \
            (context, act_property.propagated, exp_property.propagated)
        self.assert_CIMQualifier_dict(
            act_property.qualifiers, exp_property.qualifiers, context)

    def assert_CIMMethod_obj(self, act_method, exp_method, context):
        """
        Verify that two CIMMethod objects are equal.
        """

        assert act_method.name == exp_method.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_method.name, exp_method.name)
        assert act_method.return_type == exp_method.return_type, \
            "%s, return_type attribute: %s (expected: %s)" % \
            (context, act_method.return_type, exp_method.return_type)
        assert act_method.class_origin == exp_method.class_origin, \
            "%s, class_origin attribute: %s (expected: %s)" % \
            (context, act_method.class_origin, exp_method.class_origin)
        assert act_method.propagated == exp_method.propagated, \
            "%s, propagated attribute: %s methodmethod (expected: %s)" % \
            (context, act_method.propagated, exp_method.propagated)
        self.assert_CIMParameter_dict(
            act_method.parameters, exp_method.parameters, context)
        self.assert_CIMQualifier_dict(
            act_method.qualifiers, exp_method.qualifiers, context)

    def assert_CIMParameter_obj(self, act_parameter, exp_parameter, context):
        """
        Verify that two CIMParameter objects are equal.
        """

        assert act_parameter.name == exp_parameter.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_parameter.name, exp_parameter.name)
        assert act_parameter.type == exp_parameter.type, \
            "%s, type attribute: %s (expected: %s)" % \
            (context, act_parameter.type, exp_parameter.type)
        assert act_parameter.reference_class == exp_parameter.reference_class, \
            "%s, reference_class attribute: %s (expected: %s)" % \
            (context,
             act_parameter.reference_class, exp_parameter.reference_class)
        assert act_parameter.is_array == exp_parameter.is_array, \
            "%s, is_array attribute: %s (expected: %s)" % \
            (context, act_parameter.is_array, exp_parameter.is_array)
        assert act_parameter.array_size == exp_parameter.array_size, \
            "%s, array_size attribute: %s (expected: %s)" % \
            (context, act_parameter.array_size, exp_parameter.array_size)
        self.assert_CIMQualifier_dict(
            act_parameter.qualifiers, exp_parameter.qualifiers, context)

    @staticmethod
    def assert_CIMQualifier_obj(act_qualifier, exp_qualifier, context):
        """
        Verify that two CIMQualifier objects are equal.
        """

        assert act_qualifier.name == exp_qualifier.name, \
            "%s (dict key), name attribute: %s (expected: %s)" % \
            (context, act_qualifier.name, exp_qualifier.name)
        assert act_qualifier.value == exp_qualifier.value, \
            "%s, value attribute: %s (expected: %s)" % \
            (context, act_qualifier.value, exp_qualifier.value)
        assert act_qualifier.type == exp_qualifier.type, \
            "%s, type attribute: %s (expected: %s)" % \
            (context, act_qualifier.type, exp_qualifier.type)
        assert act_qualifier.propagated == exp_qualifier.propagated, \
            "%s, propagated attribute: %s (expected: %s)" % \
            (context, act_qualifier.propagated, exp_qualifier.propagated)
        assert act_qualifier.overridable == exp_qualifier.overridable, \
            "%s, overridable attribute: %s (expected: %s)" % \
            (context, act_qualifier.overridable, exp_qualifier.overridable)
        assert act_qualifier.tosubclass == exp_qualifier.tosubclass, \
            "%s, tosubclass attribute: %s (expected: %s)" % \
            (context, act_qualifier.tosubclass, exp_qualifier.tosubclass)
        assert act_qualifier.toinstance == exp_qualifier.toinstance, \
            "%s, toinstance attribute: %s (expected: %s)" % \
            (context, act_qualifier.toinstance, exp_qualifier.toinstance)
        assert act_qualifier.translatable == exp_qualifier.translatable, \
            "%s, translatable attribute: %s (expected: %s)" % \
            (context, act_qualifier.translatable, exp_qualifier.translatable)

    def assert_dict_keys(self, act_dict, exp_dict, context):
        """
        Verify that two dicts have the same set of keys.
        """
        act_keys = set(act_dict.keys())
        exp_keys = set(exp_dict.keys())
        if act_keys != exp_keys:
            missing_keys = exp_keys - act_keys
            added_keys = act_keys - exp_keys
            msg = "Different sets of %s" % context
            if missing_keys:
                msg += ", missing: %s" % list(missing_keys)
            if added_keys:
                msg += ", added: %s" % list(added_keys)
            self.fail(msg)



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
The following exceptions are pywbem specific exceptions that can be raised at
the WBEM client library API.
"""

import six
from ._cim_constants import _statuscode2name, _statuscode2string

# This module is meant to be safe for 'import *'.

__all__ = ['Error', 'ConnectionError', 'AuthError', 'HTTPError', 'TimeoutError',
           'VersionError', 'ParseError', 'CIMXMLParseError', 'XMLParseError',
           'HeaderParseError', 'CIMError', 'ModelError',
           '_RequestExceptionMixin', '_ResponseExceptionMixin']


class _RequestExceptionMixin(object):
    # pylint: disable=too-few-public-methods
    """
    An internal mixin class for pywbem specific exceptions that provides the
    ability to store the CIM-XML request string in the exception.

    *New in pywbem 1.0.*

    Derived classes using this mixin class need to specify it before the base
    error class.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters:

          *args :
            Any other positional arguments are passed on to the next superclass.

          **kwargs :
            Any other keyword arguments are passed on to the next superclass.

          request_data (:term:`string`):
            CIM-XML request string. Omitted or `None` means the exception does
            not store a CIM-XML request.
            Must be specified as a keyword argument; if specified it will be
            removed from the keyword arguments that are passed on.
        """
        if 'request_data' in kwargs:
            request_data = kwargs['request_data']
            del kwargs['request_data']
        else:
            request_data = None
        self.request_data = request_data
        super(_RequestExceptionMixin, self).__init__(*args, **kwargs)

    @property
    def request_data(self):
        """
        :term:`string`: CIM-XML request string (settable).
        `None` if the exception does not store a CIM-XML request.
        """
        return self._request_data

    @request_data.setter
    def request_data(self, request_data):
        """Setter method; for a description see the getter method."""
        self._request_data = request_data


class _ResponseExceptionMixin(object):
    # pylint: disable=too-few-public-methods
    """
    Mixin class into pywbem specific exceptions that provides the ability to
    store the CIM-XML response string in the exception.

    *New in pywbem 1.0.*

    Derived classes using this mixin class need to specify it before the base
    error class.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters:

          *args :
            Any other positional arguments are passed on to the next superclass.

          **kwargs :
            Any other keyword arguments are passed on to the next superclass.

          response_data (:term:`string`):
            CIM-XML response string. Omitted or `None` means the exception does
            not store a CIM-XML response.
            Must be specified as a keyword argument; if specified it will be
            removed from the keyword arguments that are passed on.
        """
        if 'response_data' in kwargs:
            response_data = kwargs['response_data']
            del kwargs['response_data']
        else:
            response_data = None
        self.response_data = response_data
        super(_ResponseExceptionMixin, self).__init__(*args, **kwargs)

    @property
    def response_data(self):
        """
        :term:`string`: CIM-XML response string (settable).
        `None` if the exception does not store a CIM-XML response.
        """
        return self._response_data

    @response_data.setter
    def response_data(self, response_data):
        """Setter method; for a description see the getter method."""
        self._response_data = response_data


class Error(Exception):
    """
    Abstract base class for pywbem specific exceptions.
    """

    def __init__(self, *args, **kwargs):
        """
        Parameters:

          *args :
            Any other positional arguments are passed to :exc:`py:Exception`.

          conn_id (:term:`connection id`):
            Connection ID of the connection in whose context the error
            happened. Omitted or `None` if the error did not happen in context
            of any connection, or if the connection context was not known.
            Must be specified as a keyword argument.
        """
        if 'conn_id' in kwargs:
            conn_id = kwargs['conn_id']
            del kwargs['conn_id']
        else:
            conn_id = None
        self._conn_id = conn_id
        # The Python Exception class cannot be initialized with keyword args
        assert not kwargs, str(kwargs)
        super(Error, self).__init__(*args)

    @property
    def conn_id(self):
        """
        :term:`connection id`: Connection ID of the connection in whose context
        the error happened.

        `None` if the error did not happen in context
        of any connection, or if the connection context was not known.
        """
        return self._conn_id

    @property
    def conn_str(self):
        """
        :term:`unicode string`: String that identifies the connection in
        exception messages.
        """
        ret_str = "Connection id: {0}".format(self.conn_id)
        return ret_str


class ConnectionError(Error):
    # pylint: disable=redefined-builtin
    """
    This exception indicates a problem with the connection to the WBEM
    server. A retry may or may not succeed.
    Derived from :exc:`~pywbem.Error`.
    """

    def __init__(self, message, conn_id=None):
        """
        Parameters:

          message (:term:`string`): Error message (will be put into `args[0]`).

          conn_id (:term:`connection id`):
            Connection ID of the connection in whose context the error happened.
            Omitted or `None` if the error did not happen in context of any
            connection, or if the connection context was not known.

        :ivar args: A tuple (message, ) set from the corresponding init
            arguments.
        """
        assert message is None or isinstance(message, six.string_types), \
            str(type(message))
        super(ConnectionError, self).__init__(message, conn_id=conn_id)


class AuthError(Error):
    """
    This exception indicates an authentication error with the WBEM server,
    either during TLS/SSL handshake, or during HTTP-level authentication.
    Derived from :exc:`~pywbem.Error`.
    """

    def __init__(self, message, conn_id=None):
        """
        Parameters:

          message (:term:`string`): Error message (will be put into `args[0]`).

          conn_id (:term:`connection id`):
            Connection ID of the connection in whose context the error happened.
            Omitted or `None` if the error did not happen in context of any
            connection, or if the connection context was not known.

        :ivar args: A tuple (message, ) set from the corresponding init
            arguments.
        """
        assert message is None or isinstance(message, six.string_types), \
            str(type(message))
        super(AuthError, self).__init__(message, conn_id=conn_id)


class HTTPError(_RequestExceptionMixin, _ResponseExceptionMixin, Error):
    """
    This exception indicates that the WBEM server returned an HTTP response
    with a bad HTTP status code. Derived from :exc:`~pywbem.Error`.

    The `args` instance variable is a `tuple (status, reason, cimerror,
    cimdetails)`.

    The `message` instance variable is not set.

    Occurrence of this exception nearly always indicates an issue with the
    WBEM server.
    """

    def __init__(self, status, reason, cimerror=None, cimdetails=None,
                 conn_id=None, request_data=None, response_data=None):
        """
        Parameters:

          status (:term:`integer`): HTTP status code (e.g. 500).

          reason (:term:`string`): HTTP reason phrase (e.g.
            "Internal Server Error").

          cimerror (:term:`string`): Value of the `CIMError` HTTP header field,
            if present. `None`, otherwise.

          cimdetails (dict): Dictionary with CIMOM-specific header
            fields with details about the situation reported in the `CIMError`
            header field.

            * Key: header field name (e.g. `PGErrorDetail`)
            * Value: header field value (i.e. text message)

            Passing `None` will result in an empty dictionary.

          conn_id (:term:`connection id`): Connection ID of the connection in
            whose context the error happened. `None` if the error did not
            happen in context of any connection, or if the connection context
            was not known.

          request_data (:term:`string`): CIM-XML request string.
            `None` means the exception does not store a CIM-XML request.

          response_data (:term:`string`): CIM-XML response string.
            `None` means the exception does not store a CIM-XML response.

        :ivar args: A tuple (status, reason, cimerror, cimdetails) set from the
            corresponding init arguments.
        """
        if cimdetails is None:
            cimdetails = {}
        super(HTTPError, self).__init__(
            status, reason, cimerror, cimdetails, conn_id=conn_id,
            request_data=request_data, response_data=response_data)

    @property
    def status(self):
        """
        :term:`integer`: HTTP status code.

        Example: 500

        See :term:`RFC2616` for a list of HTTP status codes and reason phrases.
        """
        # Note: The reporting of unsubscriptable-object seems to be a false
        # positive in Pylint, see https://github.com/PyCQA/pylint/issues/1498
        return self.args[0]  # pylint: disable=unsubscriptable-object

    @property
    def reason(self):
        """
        :term:`string`: HTTP reason phrase.

        Example: "Internal Server Error"

        See :term:`RFC2616` for a list of HTTP status codes and reason phrases.
        """
        return self.args[1]  # pylint: disable=unsubscriptable-object

    @property
    def cimerror(self):
        """
        :term:`string`: Value of `CIMError` HTTP header field in response, if
        present.

        `None`, otherwise.

        See :term:`DSP0200` for a list of values.
        """
        return self.args[2]  # pylint: disable=unsubscriptable-object

    @property
    def cimdetails(self):
        """
        dict: CIMOM-specific details on the situation reported in the `CIMError`
        header field.

        The value is a dictionary with:

        * Key: header field name (e.g. `PGErrorDetail`).
        * Value: header field value.
        """
        return self.args[3]  # pylint: disable=unsubscriptable-object

    def __str__(self):
        ret_str = "{0} ({1})".format(self.status, self.reason)
        if self.cimerror is not None:
            ret_str += ", CIMError: {0}".format(self.cimerror)
        for key in self.cimdetails:
            ret_str += ", {0}: {1}".format(key, self.cimdetails[key])
        return ret_str


class TimeoutError(Error):
    # pylint: disable=redefined-builtin
    """
    This exception indicates that the client timed out waiting for the WBEM
    server. Derived from :exc:`~pywbem.Error`.
    """

    def __init__(self, message, conn_id=None):
        """
        Parameters:

          message (:term:`string`): Error message (will be put into `args[0]`).

          conn_id (:term:`connection id`):
            Connection ID of the connection in whose context the error happened.
            Omitted or `None` if the error did not happen in context of any
            connection, or if the connection context was not known.

        :ivar args: A tuple (message, ) set from the corresponding init
            arguments.
        """
        assert message is None or isinstance(message, six.string_types), \
            str(type(message))
        super(TimeoutError, self).__init__(message, conn_id=conn_id)


class ParseError(_RequestExceptionMixin, _ResponseExceptionMixin, Error):
    """
    This exception indicates a parsing error with the CIM-XML operation
    response the pywbem client received, or with the CIM-XML indication
    request the pywbem listener received.

    Derived from :exc:`~pywbem.Error`.

    The CIM-XML response data is part of the `str()` representation of the
    exception.

    This exception is a base class for more specific exceptions:

    * :exc:`~pywbem.CIMXMLParseError` - Issue at the CIM-XML level (e.g. NAME
      attribute missing on CLASS element)
    * :exc:`~pywbem.XMLParseError` - Issue at the XML level (e.g. ill-formed
      XML)
    * :exc:`~pywbem.HeaderParseError` - Issue with HTTP headers (e.g. invalid
      content-type header)

    Occurrence of this exception nearly always indicates an issue with the
    WBEM server.
    """

    def __init__(self, message, conn_id=None, request_data=None,
                 response_data=None):
        """
        Parameters:

          message (:term:`string`): Error message (will be put into `args[0]`).

          conn_id (:term:`connection id`): Connection ID of the connection in
            whose context the error happened. `None` if the error did not
            happen in context of any connection, or if the connection context
            was not known.

          request_data (:term:`string`): CIM-XML request string.
            `None` means the exception does not store a CIM-XML request.

          response_data (:term:`string`): CIM-XML response string.
            `None` means the exception does not store a CIM-XML response.

        :ivar args: A tuple (message, ) set from the corresponding init
            argument.
        """
        assert message is None or isinstance(message, six.string_types), \
            str(type(message))
        super(ParseError, self).__init__(
            message, conn_id=conn_id, request_data=request_data,
            response_data=response_data)

    def __str__(self):
        error_str = super(ParseError, self).__str__()
        ret_str = "{0}\nCIM-XML response: {1}". \
            format(error_str, self.response_data)
        return ret_str


class CIMXMLParseError(ParseError):
    """
    This exception indicates a specific kind of :exc:`~pywbem.ParseError`
    that is an issue at the CIM-XML level.

    Example: 'NAME' attribute missing on 'CLASS' element.

    Occurrence of this exception nearly always indicates an issue with the
    WBEM server.
    """
    pass


class XMLParseError(ParseError):
    """
    This exception indicates a specific kind of :exc:`~pywbem.ParseError`
    that is an issue at the XML level.

    Example: Ill-formed XML.

    Occurrence of this exception nearly always indicates an issue with the
    WBEM server.
    """
    pass


class HeaderParseError(ParseError):
    """
    This exception indicates a specific kind of :exc:`~pywbem.ParseError`
    that is an issue with an HTTP header.

    Example: Invalid content-type.

    Occurrence of this exception nearly always indicates an issue with the
    WBEM server.
    """
    pass


class VersionError(Error):
    """
    This exception indicates an unsupported CIM, DTD or protocol version
    with the CIM-XML response returned by the WBEM server, or in the CIM-XML
    request sent by the WBEM listener. Derived from :exc:`~pywbem.Error`.
    """

    def __init__(self, message, conn_id=None):
        """
        Parameters:

          message (:term:`string`): Error message (will be put into `args[0]`).

          conn_id (:term:`connection id`):
            Connection ID of the connection in whose context the error happened.
            Omitted or `None` if the error did not happen in context of any
            connection, or if the connection context was not known.

        :ivar args: A tuple (message, ) set from the corresponding init
            argument.
        """
        assert message is None or isinstance(message, six.string_types), \
            str(type(message))
        super(VersionError, self).__init__(message, conn_id=conn_id)


class CIMError(_RequestExceptionMixin, Error):
    """
    This exception indicates that the WBEM server returned an error response
    with a CIM status code. Derived from :exc:`~pywbem.Error`.

    Accessing the CIM status code of a :class:`CIMError` object:

      In Python 2, any :exc:`~py:exceptions.Exception` object can be accessed
      by index and slice and will delegate such access to its
      :attr:`~py:exceptions.BaseException.args` instance variable. In Python 3,
      that ability has been removed.

      In pywbem 0.9, the :attr:`~pywbem.CIMError.status_code` and
      :attr:`~pywbem.CIMError.status_description` properties were added.

      Therefore, the following approach is not recommended, because it does not
      work on Python 3::

          except CIMError as exc:
              status_code = exc[0]

      The following approach works for pywbem 0.7 or newer::

          except CIMError as exc:
              status_code = exc.args[0]

      The following approach is recommended when using pywbem 0.9 or newer::

          except CIMError as exc:
              status_code = exc.status_code
    """

    # pylint: disable=super-init-not-called
    def __init__(self, status_code, status_description=None, instances=None,
                 conn_id=None, request_data=None):
        """
        Parameters:

          status_code (:term:`integer`): Numeric CIM status code.

          status_description (:term:`string`): CIM status description text
            returned by the server, representing a human readable message
            describing the error. `None`, if the server did not return
            a description text.

          instances (list of :class:`~pywbem.CIMInstance`): List of CIM
            instances returned by the WBEM server in the error response, that
            provide more details on the error. `None` if there are no such
            instances.

          conn_id (:term:`connection id`): Connection ID of the connection in
            whose context the error happened. `None` if the error did not
            happen in context of any connection, or if the connection context
            was not known.

          request_data (:term:`string`): CIM-XML request string.
            `None` means the exception does not store a CIM-XML request.

        :ivar args: A tuple (status_code, status_description, instances) set
            from the corresponding init arguments.
        """
        super(CIMError, self).__init__(
            status_code, status_description, instances, conn_id=conn_id,
            request_data=request_data)

    @property
    def status_code(self):
        """
        :term:`integer`: Numeric CIM status code.

        *New in pywbem 0.9.*

        Example: 5

        See :ref:`CIM status codes` for constants defining the numeric CIM
        status code values."""
        return self.args[0]  # pylint: disable=unsubscriptable-object

    @property
    def status_code_name(self):
        """
        :term:`string`: Symbolic name of the CIM status code.

        Example: "CIM_ERR_INVALID_CLASS"

        *New in pywbem 0.9.*

        If the CIM status code is invalid, the string
        "Invalid status code <status_code>" is returned."""
        return _statuscode2name(self.status_code)

    @property
    def status_description(self):
        """
        :term:`string`: CIM status description text returned by the server,
        representing a human readable message describing the error.

        *New in pywbem 0.9.*

        Example: "The specified class does not exist."

        If the server did not return a description, a short default text for
        the CIM status code is returned. If the CIM status code is invalid,
        the string "Invalid status code <status_code>" is returned.
        """
        desc = self.args[1]  # pylint: disable=unsubscriptable-object
        return desc or _statuscode2string(self.status_code)

    @property
    def instances(self):
        """
        List of :class:`~pywbem.CIMInstance`: CIM instances returned by the
        WBEM server in the error response, that provide more details on the
        error.

        *New in pywbem 0.13.*

        `None` if there are no such instances.
        """
        return self.args[2]  # pylint: disable=unsubscriptable-object

    def __str__(self):
        inst_str = " ({0} instances)".format(len(self.instances)) \
            if self.instances else ""
        ret_str = "{0} ({1}): {2}{3}".format(
            self.status_code, self.status_code_name, self.status_description,
            inst_str)
        return ret_str


class ModelError(Error):
    """
    This exception indicates an error with the model implemented by the WBEM
    server, that was detected by the pywbem client.

    Examples are mismatches in data types of CIM elements (properties, methods,
    parameters) between classes and instances, CIM elements that appear in
    instances without being declared in classes, or violations of requirements
    defined in advertised management profiles.

    Derived from :exc:`~pywbem.Error`.
    """

    def __init__(self, message, conn_id=None):
        """
        Parameters:

          message (:term:`string`): Error message (will be put into `args[0]`).

          conn_id (:term:`connection id`):
            Connection ID of the connection in whose context the error happened.
            Omitted or `None` if the error did not happen in context of any
            connection, or if the connection context was not known.

        :ivar args: A tuple (message, ) set from the corresponding init
            argument.
        """
        assert message is None or isinstance(message, six.string_types), \
            str(type(message))
        super(ModelError, self).__init__(message, conn_id=conn_id)
